# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

__all__ = [
    "_add_output_manifests_from_s3",
    "_download_all_manifests_with_absolute_paths",
    "_merge_absolute_path_manifest_list",
    "_download_manifest_paths",
]

from typing import Any, Callable, DefaultDict, Optional
from datetime import datetime, timezone
import re
import os
import concurrent.futures
from threading import Lock
from pathlib import Path
import posixpath
import ntpath

import boto3
from boto3.s3.transfer import ProgressCallbackInvoker
from botocore.client import BaseClient
from botocore.exceptions import BotoCoreError, ClientError

from ..download import (
    _get_output_manifest_prefix,
    _get_tasks_manifests_keys_from_s3,
    _get_asset_root_and_manifest_from_s3_with_last_modified,
    _get_num_download_workers,
    _get_new_copy_file_path,
    S3_DOWNLOAD_MAX_CONCURRENCY,
)
from ..asset_manifests import (
    hash_data as ja_hash_data,
    BaseAssetManifest,
    BaseManifestPath,
    HashAlgorithm,
)
from ..asset_manifests.v2023_03_03.asset_manifest import DEFAULT_HASH_ALG
from ..models import (
    FileConflictResolution,
    JobAttachmentS3Settings,
    S3_MANIFEST_FOLDER_NAME,
    PathFormat,
)
from ..exceptions import (
    COMMON_ERROR_GUIDANCE_FOR_S3,
    AssetSyncError,
    AssetSyncCancelledError,
    JobAttachmentS3BotoCoreError,
    JobAttachmentsS3ClientError,
    JobAttachmentsError,
)
from .._aws.aws_clients import (
    get_account_id,
    get_s3_client,
    get_s3_transfer_manager,
)
from ..progress_tracker import (
    ProgressTracker,
    ProgressStatus,
    ProgressReportMetadata,
)
from .._utils import _get_long_path_compatible_path
from ...job_attachments._path_mapping import _PathMappingRuleApplier


"""
This file contains a forked copy of some functionality from deadline.job_attachments.download,
with its interface refactored to support the incremental download command.

It consists fully of internal-only functionality. We would like to iteratively refine these
interfaces over time, with the goal that we deprecate the current interfaces
in deadline.job_attachments.download and replace them with more general and flexible interfaces.
"""

SESSION_ACTION_ID_FROM_KEY_RE = re.compile(r"(sessionaction-[^/-]+-[^/-]+)/")


def _add_output_manifests_from_s3(
    farm_id: str,
    queue: dict[str, Any],
    job: dict[str, Any],
    boto3_session: boto3.Session,
    session_action_list: list[dict[str, Any]],
):
    """
    This function takes a list of session actions (as returned by boto3 deadline.list_session_actions),
    and for any that lack manifest fields, updates them with values retrieved from S3. While the response
    from Deadline Cloud will always return both outputManifestPath and outputManifestHash, this function
    only populates the outputManifestPath value. The order of the manifests in the list precisely correspond
    to the manifests returned by boto3 deadline.get_job, clients of these APIs can zip() the two
    manifests lists together to get the full set of fields needed for processing.

    * https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/deadline/client/list_session_actions.html
    * https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/deadline/client/get_job.html

    Args:
        farm_id: The farm id.
        queue: The queue as returned by boto3 deadline.get_queue().
        job: The job as returned by boto3 deadline.search_jobs().
        boto3_session: The boto3.Session for accessing AWS.
        session_action_list: A list of session actions to modify by adding the "manifests" field where necessary.
            Its shape is as returned by boto3 deadline.list_session_actions() or deadline.get_session_action().
    """
    # If the job has no job attachments, there's nothing to add
    if "attachments" not in job:
        return

    s3_settings = JobAttachmentS3Settings(**queue["jobAttachmentSettings"])

    # Filter the session action list to exclude any that already contain manifests,
    # then return early if the result is empty.
    session_action_list = [
        session_action
        for session_action in session_action_list
        if "manifests" not in session_action
    ]
    if not session_action_list:
        return

    # Process the job's attachments list to generate the hashes
    job_manifests_length = len(job["attachments"]["manifests"])
    job_indexed_root_path_hash = [
        (
            index,
            ja_hash_data(
                f"{manifest.get('fileSystemLocationName', '')}{manifest['rootPath']}".encode(),
                DEFAULT_HASH_ALG,
            ),
        )
        for index, manifest in enumerate(job["attachments"]["manifests"])
    ]
    # Initialize to empty "manifests" entries
    for session_action in session_action_list:
        session_action["manifests"] = [{}] * job_manifests_length

    # Get all the output manifest keys for all the steps/tasks of the job.
    # TODO: This is very inefficient if the job has lots of tasks, because
    #       the incremental download will generally only use a few of them at a time.
    manifest_prefix: str = _get_output_manifest_prefix(
        s3_settings, farm_id, queue["queueId"], job["jobId"]
    )
    try:
        manifests_keys: list[str] = _get_tasks_manifests_keys_from_s3(
            manifest_prefix,
            s3_settings.s3BucketName,
            session=boto3_session,
            select_latest_per_task=False,
        )
    except JobAttachmentsError as e:
        # If there are no manifests, treat as no data.
        if str(e).startswith("Unable to find asset manifest in"):
            return
        else:
            raise

    # Organize the session actions by session action id, so that we can quickly get
    # to the correct session action from the manifest object key.
    session_actions_by_session_action_id: dict[str, dict[str, Any]] = {
        session_action["sessionActionId"]: session_action for session_action in session_action_list
    }

    manifest_prefix = f"{queue['jobAttachmentSettings']['rootPrefix']}/{S3_MANIFEST_FOLDER_NAME}/"
    for key in manifests_keys:
        # Extract the session action id from the manifest key
        m = SESSION_ACTION_ID_FROM_KEY_RE.search(key)
        if m:
            manifest_session_action_id = m[1]
        else:
            raise RuntimeError(
                f"Job attachments manifest key for job {job['name']} ({job['jobId']}) lacks a session action id"
            )
        # Loop through all the manifests to see whether the hash of the rootPath is in the key,
        # in order to determine which position in the manifests list this key should get set in.
        manifests_index = None
        for index, root_path_hash in job_indexed_root_path_hash:
            if root_path_hash in key:
                manifests_index = index
                break
        if manifests_index is None:
            root_path_hashes = [hash for _, hash in job_indexed_root_path_hash]
            raise RuntimeError(
                f"Job attachments manifest key for job {job['name']} ({job['jobId']}) does not contain any of the rootPath hashes {', '.join(root_path_hashes)}: {key}"
            )
        # If this session action is in the list, add this key to it
        session_action_for_key = session_actions_by_session_action_id.get(
            manifest_session_action_id
        )
        if session_action_for_key is not None:
            # This is equivalent to "output_manifest_path = key.removeprefix(manifest_prefix)" to
            # retain support for Python 3.8 which does not support str.removeprefix.
            output_manifest_path = key
            if output_manifest_path.startswith(manifest_prefix):
                output_manifest_path = output_manifest_path[len(manifest_prefix) :]
            session_action_for_key["manifests"][manifests_index] = {
                "outputManifestPath": output_manifest_path,
            }


def _download_manifest_and_make_paths_absolute(
    index: int,
    queue: dict[str, Any],
    job_id: str,
    root_path: str,
    manifest_s3_key: str,
    path_mapping_rule_applier: Optional[_PathMappingRuleApplier],
    boto3_session_for_s3: boto3.Session,
    output_manifests: list,
    output_unmapped_paths: list[tuple[str, str]],
):
    """
    Downloads the specified manifest, makes all its paths absolute using root_path,
    and then places it in output_manifests[index].
    """
    # Download the manifest
    _, last_modified, manifest = _get_asset_root_and_manifest_from_s3_with_last_modified(
        manifest_s3_key, queue["jobAttachmentSettings"]["s3BucketName"], boto3_session_for_s3
    )
    if path_mapping_rule_applier:
        new_manifest_paths = []
        if path_mapping_rule_applier.source_path_format == PathFormat.WINDOWS.value:
            source_os_path: Any = ntpath
        else:
            source_os_path = posixpath
    else:
        source_os_path = os.path

    # Convert all the manifest paths to have absolute normalized local paths
    for manifest_path in manifest.paths:
        manifest_path.path = source_os_path.normpath(
            source_os_path.join(root_path, manifest_path.path)
        )
        if path_mapping_rule_applier:
            try:
                manifest_path.path = str(
                    path_mapping_rule_applier.strict_transform(manifest_path.path)
                )
                new_manifest_paths.append(manifest_path)
            except ValueError:
                output_unmapped_paths.append((job_id, manifest_path.path))

    if path_mapping_rule_applier:
        # Update the manifest to only include the mapped paths
        manifest.paths = new_manifest_paths

    output_manifests[index] = (last_modified, manifest)


def _get_manifests_to_download(
    job_attachments_root_prefix: str,
    download_candidate_jobs: dict[str, dict[str, Any]],
    job_sessions: dict[str, list],
    path_mapping_rule_appliers: dict[str, Optional[_PathMappingRuleApplier]],
) -> list[tuple[Optional[_PathMappingRuleApplier], str, str, str]]:
    """
    Collect a list of (pathMappingRuleApplier, jobId, rootPath, manifest_s3_key) tuples for all the job attachments that need to be downloaded.

    Args:
        job_attachments_root_prefix: The queue.jobAttachmentSettings.rootPrefix field from the Deadline
            Cloud queue.
        download_candidate_jobs: A mapping from job id to jobs as returned by deadline.search_jobs.
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
        path_mapping_rule_appliers: A mapping from storage profile ID to the path mapping rule applier to use for it.
            If no path mapping should be used, is the empty {}.

    Returns:
        A list of (pathMappingRuleApplier, jobId, rootPath, manifest_s3_key) tuples for the manifest objects that need to be downloaded.
    """
    manifests_to_download: list[tuple[Optional[_PathMappingRuleApplier], str, str, str]] = []
    for job_id, session_list in job_sessions.items():
        job = download_candidate_jobs[job_id]
        for session in session_list:
            for session_action in session.get("sessionActions", []):
                # The manifests lists from the job and session action correspond, so we can zip them
                # together to attach the root path with the S3 manifest key
                for job_manifest, session_action_manifest in zip(
                    job["attachments"]["manifests"], session_action["manifests"]
                ):
                    if "outputManifestPath" in session_action_manifest:
                        manifests_to_download.append(
                            (
                                path_mapping_rule_appliers[job["storageProfileId"]]
                                if path_mapping_rule_appliers
                                else None,
                                job_id,
                                job_manifest["rootPath"],
                                "/".join(
                                    [
                                        job_attachments_root_prefix,
                                        S3_MANIFEST_FOLDER_NAME,
                                        session_action_manifest["outputManifestPath"],
                                    ]
                                ),
                            )
                        )
    return manifests_to_download


def _download_all_manifests_with_absolute_paths(
    queue: dict[str, Any],
    download_candidate_jobs: dict[str, dict[str, Any]],
    job_sessions: dict[str, list],
    path_mapping_rule_appliers: dict[str, Optional[_PathMappingRuleApplier]],
    output_unmapped_paths: dict[str, list[str]],
    boto3_session_for_s3: boto3.Session,
    print_function_callback: Callable[[str], None] = lambda msg: None,
) -> list[tuple[datetime, BaseAssetManifest]]:
    """
    Downloads all the manifest files that are in the session_actions of job_sessions, and uses the rootPath
    values taken from the job to make all the paths in the manifest absolute.

    Args:
        queue: The Deadline Cloud queue as returned by deadline.get_queue.
        download_candidate_jobs: A mapping from job id to jobs as returned by deadline.search_jobs.
        job_sessions: Contains each job's sessions and session actions, structured as job_sessions[job_id][session_index]["sessionActions"][session_action_index].
                      See the function _get_job_sessions for more details.
        path_mapping_rule_appliers: A mapping from storage profile ID to the path mapping rule applier to use for it.
            If no path mapping should be used, is the empty {}.
        boto3_session_for_s3: The boto3.Session to use for accessing S3.
        output_unmapped_paths: A mapping from the job id to a list of all the paths that were not mapped
            and therefore will not be downloaded.
        print_function_callback: Callback for printing output to the terminal or log.

    Returns:
        A list of BaseAssetManifest objects containing local absolute file paths sorted by the last_modified timestamp.
    """
    # Get the list of (rootPath, manifest_s3_key) tuples to download from S3.
    manifests_to_download: list[tuple[Optional[_PathMappingRuleApplier], str, str, str]] = (
        _get_manifests_to_download(
            queue["jobAttachmentSettings"]["rootPrefix"],
            download_candidate_jobs,
            job_sessions,
            path_mapping_rule_appliers,
        )
    )

    print_function_callback("")
    print_function_callback(f"Downloading {len(manifests_to_download)} asset manifests from S3...")
    start_time = datetime.now(tz=timezone.utc)

    # Download all the manifest files from S3, and make the paths in the manifests absolute local paths
    # by joining with the root path and normalizing
    downloaded_manifests: list = [None] * len(manifests_to_download)
    # All the unmapped paths get recorded here as (jobId, unmappedPath)
    unmapped_paths: list[tuple[str, str]] = []

    max_workers = S3_DOWNLOAD_MAX_CONCURRENCY
    print_function_callback(f"Using {max_workers} threads")
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for index, (path_mapping_rule_applier, job_id, root_path, manifest_s3_key) in enumerate(
            manifests_to_download
        ):
            futures.append(
                executor.submit(
                    _download_manifest_and_make_paths_absolute,
                    index,
                    queue,
                    job_id,
                    root_path,
                    manifest_s3_key,
                    path_mapping_rule_applier,
                    boto3_session_for_s3,
                    downloaded_manifests,
                    unmapped_paths,
                )
            )
        # surfaces any exceptions in the thread
        for future in concurrent.futures.as_completed(futures):
            future.result()

    # Transform the unmapped paths into the output, grouping by job id
    for job_id, unmapped_path in unmapped_paths:
        output_unmapped_paths.setdefault(job_id, []).append(unmapped_path)

    duration = datetime.now(tz=timezone.utc) - start_time
    print_function_callback(f"...downloaded manifests in {duration}")

    return downloaded_manifests


def _merge_absolute_path_manifest_list(
    downloaded_manifests: list[tuple[datetime, BaseAssetManifest]],
) -> list[BaseManifestPath]:
    """
    Given a list of manifests that contain absolute paths, uses the provided last modified timestamps
    to sort them, and merges them all into a single manifest. Returns the list of manifest paths
    for download.

    Args:
        downloaded_manifests: A list of (last_modified_timestamp, manifest) tuples, where each last
            modified timestamp is the LastModified datetime from the S3 object holding the manifest.

    Returns:
        A list of manifest paths to download, the result of merging the manifests.
    """
    # Because the paths in manifests are all absolute and normalized now, we can merge them
    # in order by inserting them into a dict in order, using the normcased path as the key.
    # Later files of the same name will overwrite earlier ones.

    # Sort the manifests by last modified, so that we can overlay them
    # with later manifests overwriting files from earlier ones.
    downloaded_manifests.sort(key=lambda item: item[0])

    merged_manifest_paths_dict = {}
    for _, manifest in downloaded_manifests:
        for manifest_path in manifest.paths:
            merged_manifest_paths_dict[os.path.normcase(manifest_path.path)] = manifest_path
    return list(merged_manifest_paths_dict.values())


def _download_file_with_transfer_manager(
    local_file_path: Path,
    s3_bucket: str,
    s3_key: str,
    boto3_session: boto3.Session,
    s3_client: BaseClient,
    progress_tracker: ProgressTracker,
):
    """
    Downloads a single file from S3 using S3 transfer manager. This is appropriate for a larger
    file that benefits from parallel multi-part download.
    """
    transfer_manager = get_s3_transfer_manager(s3_client=s3_client)

    future: concurrent.futures.Future

    def handler(bytes_downloaded):
        nonlocal progress_tracker
        nonlocal future

        should_continue = progress_tracker.track_progress_callback(bytes_downloaded)
        if not should_continue:
            future.cancel()

    subscribers = [ProgressCallbackInvoker(handler)]

    future = transfer_manager.download(
        bucket=s3_bucket,
        key=s3_key,
        fileobj=str(local_file_path),
        extra_args={"ExpectedBucketOwner": get_account_id(session=boto3_session)},
        subscribers=subscribers,
    )

    future.result()


def _download_file_with_get_object(
    local_file_path: Path,
    s3_bucket: str,
    s3_key: str,
    boto3_session: boto3.Session,
    s3_client: BaseClient,
    progress_tracker: ProgressTracker,
):
    """
    Downloads a single file from S3 using get_object. This is appropriate for a smaller
    file that benefits from reduced overhead.
    """
    res = s3_client.get_object(
        Bucket=s3_bucket,
        Key=s3_key,
        ExpectedBucketOwner=get_account_id(session=boto3_session),
    )
    body = res["Body"]
    # Copy the data this amount at a time
    buffer_size = 128 * 1024
    with open(local_file_path, "wb") as fh:
        while True:
            data = body.read(buffer_size)
            if not data:
                break
            should_continue = progress_tracker.track_progress_callback(len(data))
            if not should_continue:
                fh.close()
                os.remove(local_file_path)
                raise AssetSyncCancelledError("File download cancelled.")
            fh.write(data)


def _download_file(
    file: BaseManifestPath,
    hash_algorithm: HashAlgorithm,
    collision_lock: Lock,
    collision_file_dict: DefaultDict[str, int],
    s3_bucket: str,
    cas_prefix: str,
    s3_client: BaseClient,
    boto3_session_for_s3: boto3.Session,
    progress_tracker: ProgressTracker,
    file_conflict_resolution: FileConflictResolution,
) -> None:
    """
    Downloads a file from the S3 bucket to the local directory.

    Args:
        file: A BaseManifestPath whose path is a local absolute path.
        hash_algorithm: The hash algorithm used for the queue.
        collision_lock: A lock to ensure only one thread resolves a path name collision at a time.
        collision_file_dict: Dictionary for tracking path name collisions.
        s3_bucket: The job attachments S3 bucket.
        cas_prefix: The prefix for content-addressed data files in the S3 bucket.
        s3_client: A boto3 client for accessing S3.
        boto3_session_for_s3: The boto3.Session to use for accessing S3.
        progress_tracker: Object to update with download progress status.
        file_conflict_resolution: The strategy to use for file conflict resolution.
    """
    local_file_path = _get_long_path_compatible_path(file.path)

    s3_key = f"{cas_prefix}/{file.hash}.{hash_algorithm.value}"

    # If the file name already exists, resolve the conflict based on the file_conflict_resolution
    if local_file_path.is_file():
        if file_conflict_resolution == FileConflictResolution.SKIP:
            return
        elif file_conflict_resolution == FileConflictResolution.OVERWRITE:
            pass
        elif file_conflict_resolution == FileConflictResolution.CREATE_COPY:
            local_file_path = _get_new_copy_file_path(
                local_file_path, collision_lock, collision_file_dict
            )
        else:
            raise ValueError(
                f"Unknown choice for file conflict resolution: {file_conflict_resolution}"
            )

    local_file_path.parent.mkdir(parents=True, exist_ok=True)

    if file.size > 1024 * 1024:
        download_file = _download_file_with_transfer_manager
    else:
        download_file = _download_file_with_get_object

    try:
        download_file(
            local_file_path=local_file_path,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            boto3_session=boto3_session_for_s3,
            s3_client=s3_client,
            progress_tracker=progress_tracker,
        )
    except concurrent.futures.CancelledError as ce:
        if progress_tracker and progress_tracker.continue_reporting is False:
            raise AssetSyncCancelledError("File download cancelled.")
        else:
            raise AssetSyncError("File download failed.", ce) from ce
    except ClientError as exc:
        status_code = int(exc.response["ResponseMetadata"]["HTTPStatusCode"])
        status_code_guidance = {
            **COMMON_ERROR_GUIDANCE_FOR_S3,
            403: (
                (
                    "Forbidden or Access denied. Please check your AWS credentials, and ensure that "
                    "your AWS IAM Role or User has the 's3:GetObject' permission for this bucket. "
                )
                if "kms:" not in str(exc)
                else (
                    "Forbidden or Access denied. Please check your AWS credentials and Job Attachments S3 bucket "
                    "encryption settings. If a customer-managed KMS key is set, confirm that your AWS IAM Role or "
                    "User has the 'kms:Decrypt' and 'kms:DescribeKey' permissions for the key used to encrypt the bucket."
                )
            ),
            404: (
                "Not found. Please check your bucket name and object key, and ensure that they exist in the AWS account."
            ),
        }
        raise JobAttachmentsS3ClientError(
            action="downloading file",
            status_code=status_code,
            bucket_name=s3_bucket,
            key_or_prefix=s3_key,
            message=f"{status_code_guidance.get(status_code, '')} {str(exc)} (Failed to download the file to {str(local_file_path)})",
        ) from exc
    except BotoCoreError as bce:
        raise JobAttachmentS3BotoCoreError(
            action="downloading file",
            error_details=str(bce),
        ) from bce
    except Exception as e:
        raise AssetSyncError(e) from e

    # The modified time in the manifest is in microseconds, but utime requires the time be expressed in seconds.
    modified_time_override = file.mtime / 1000000  # type: ignore[attr-defined]
    os.utime(local_file_path, (modified_time_override, modified_time_override))  # type: ignore[arg-type]

    # Verify that what we downloaded has the correct file size from the manifest.
    file_size_on_disk = os.path.getsize(local_file_path)
    if file_size_on_disk != file.size:
        # TODO: Improve this error message
        raise JobAttachmentsError(
            f"File from S3 for {file.path} had incorrect size {file_size_on_disk}. Required size: {file.size}"
        )


def _download_manifest_paths(
    manifest_paths_to_download: list[BaseManifestPath],
    hash_algorithm: HashAlgorithm,
    queue: dict[str, Any],
    boto3_session_for_s3: boto3.Session,
    file_conflict_resolution: FileConflictResolution,
    on_downloading_files: Optional[Callable[[ProgressReportMetadata], bool]],
    print_function_callback: Callable[[str], None] = lambda msg: None,
) -> None:
    """
    Downloads all files from the S3 bucket in the Job Attachment settings to the specified directory.

    Args:
        manifest_paths_to_download: A list of manifest path objects to download, whose path is an absolute file system path.
        hash_algorithm: The hash algorithm in use by the queue.
        queue: The queue as returned by boto3 deadline.get_queue().
        boto3_session_for_s3: The boto3.Session to use for accessing S3.
        file_conflict_resolution: The strategy to use for file conflict resolution.
        on_downloading_files: A callback to handle progress messages and cancelation.
        print_function_callback: Callback for printing output to the terminal or log.
    """
    s3_settings = JobAttachmentS3Settings(**queue["jobAttachmentSettings"])
    s3_client = get_s3_client(session=boto3_session_for_s3)
    max_workers = _get_num_download_workers()

    collision_lock: Lock = Lock()
    collision_file_dict: DefaultDict[str, int] = DefaultDict(int)
    full_cas_prefix: str = s3_settings.full_cas_prefix()

    progress_tracker = ProgressTracker(
        status=ProgressStatus.DOWNLOAD_IN_PROGRESS,
        total_files=len(manifest_paths_to_download),
        total_bytes=sum(manifest_path.size for manifest_path in manifest_paths_to_download),
        on_progress_callback=on_downloading_files,
    )

    print_function_callback(f"Using {max_workers} threads")
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [
            executor.submit(
                _download_file,
                manifest_path,
                hash_algorithm,
                collision_lock,
                collision_file_dict,
                s3_settings.s3BucketName,
                full_cas_prefix,
                s3_client,
                boto3_session_for_s3,
                progress_tracker,
                file_conflict_resolution,
            )
            for manifest_path in manifest_paths_to_download
        ]
        # surfaces any exceptions in the thread
        for future in concurrent.futures.as_completed(futures):
            future.result()
            if progress_tracker:
                progress_tracker.increase_processed(1, 0)
                progress_tracker.report_progress()

    # to report progress 100% at the end
    if progress_tracker:
        progress_tracker.report_progress()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
API methods for queue role credentials.
"""

from configparser import ConfigParser
from typing import Dict, Any, Optional

from . import _session, _telemetry


@_telemetry.record_function_latency_telemetry_event()
def assume_queue_role_for_user(
    farmId: str, queueId: str, *, config: Optional[ConfigParser] = None
) -> Dict[str, Any]:
    """
    Assumes the user role for a queue and returns temporary credentials.

    These credentials can be used to perform user-level operations on the queue,
    such as submitting jobs and monitoring job status.

    Args:
        farmId: The ID of the farm containing the queue.
        queueId: The ID of the queue to assume the role for.
        config: Optional configuration to use. If not provided, the default configuration is used.

    Returns:
        A dictionary containing the temporary credentials in the following format:
        {
            "credentials": {
                "accessKeyId": str,
                "secretAccessKey": str,
                "sessionToken": str,
                "expiration": datetime
            }
        }

    Raises:
        ClientError: If there is an error assuming the role.
    """
    client = _session.get_boto3_client("deadline", config=config)
    response = client.assume_queue_role_for_user(farmId=farmId, queueId=queueId)
    return response


@_telemetry.record_function_latency_telemetry_event()
def assume_queue_role_for_read(
    farmId: str, queueId: str, *, config: Optional[ConfigParser] = None
) -> Dict[str, Any]:
    """
    Assumes the read role for a queue and returns temporary credentials.

    These credentials can be used to perform read-only operations on the queue,
    such as viewing job status and queue information.

    Args:
        farmId: The ID of the farm containing the queue.
        queueId: The ID of the queue to assume the role for.
        config: Optional configuration to use. If not provided, the default configuration is used.

    Returns:
        A dictionary containing the temporary credentials in the following format:
        {
            "credentials": {
                "accessKeyId": str,
                "secretAccessKey": str,
                "sessionToken": str,
                "expiration": datetime
            }
        }

    Raises:
        ClientError: If there is an error assuming the role.
    """
    client = _session.get_boto3_client("deadline", config=config)
    response = client.assume_queue_role_for_read(farmId=farmId, queueId=queueId)
    return response



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import re
import sys
import urllib.parse
from logging import getLogger
from typing import Any, Dict, List

import click

from ..exceptions import DeadlineOperationError

logger = getLogger(__name__)

__all__ = [
    "parse_query_string",
    "install_deadline_web_url_handler",
    "uninstall_deadline_web_url_handler",
    "DEADLINE_URL_SCHEME_NAME",
]

DEADLINE_URL_SCHEME_NAME = "deadline"
DEADLINE_ID_PATTERN = re.compile(r"^[0-9a-f]{32}$")
DEADLINE_TASK_ID_PATTERN = re.compile(r"^[0-9a-f]{32}-(0|([1-9][0-9]{0,9}))$")
VALID_RESOURCE_NAMES_IN_ID = ["farm", "queue", "job", "step", "task"]


def parse_query_string(
    query_string: str, parameter_names: List[str], required_parameter_names: List[str]
) -> Dict[str, str]:
    """
    Parses the URL query string into {"parameter": "value"} form.

    Any "-" in the parameter names are switched to "_" in the result.

    Args:
        query_string (str): The query string from a parsed web URL.
    """
    result: dict = {}

    if query_string:
        parsed_qs = urllib.parse.parse_qs(query_string, strict_parsing=True)
    else:
        parsed_qs = {}

    # Ensure the required parameters are provided
    missing_required_parameters = set(required_parameter_names) - set(parsed_qs.keys())
    if missing_required_parameters:
        raise DeadlineOperationError(
            f"The URL query did not contain the required parameter(s) {list(missing_required_parameters)}"
        )

    # Process all the valid parameter names
    for name in parameter_names:
        values = parsed_qs.pop(name, None)
        if values:
            if len(values) > 1:
                raise DeadlineOperationError(
                    f"The URL query parameter {name} was provided multiple times, it may only be provided once."
                )
            result[name.replace("-", "_")] = values[0]

    # If there are any left, they are not valid
    if parsed_qs:
        raise DeadlineOperationError(
            f"The URL query contained unsupported parameter names {parsed_qs.keys()}"
        )

    return result


def validate_resource_ids(ids: Dict[str, str]) -> None:
    """
    Validates that the resource IDs are all valid.
    i.e. "<name of resource>-<a hexadecimal string of length 32>"
    for Task ID: "task-<a hexadecimal string of length 32>-<0 or a number up to 10 digits long>"

    Args:
        ids (Dict[str, str]): The resource IDs to validate.
        Expected to be {"<resource type>_id": "<full id string>"} form.
    """
    for id_name, id_str in ids.items():
        resource_type = id_str.split("-")[0]
        if not id_name.startswith(resource_type) or not validate_id_format(resource_type, id_str):
            raise DeadlineOperationError(
                f'The given resource ID "{id_name}": "{id_str}" has invalid format.'
            )


def validate_id_format(resource_type: str, full_id_str: str) -> bool:
    """
    Validates if the ID is in correct format. The ID must
    - start with one of the resource names, and followed by a hyphen ("-"), and
    - the string that follows must be a hexadecimal string of length 32.
    - Additional for "task": followed by another hyphen ("-"), and ends with
      a string of either 0 or a number up to 10 digits long.

    Args:
        full_id_str (str): The ID to validate.
        resource_type (str): "farm", "queue", "job", etc.
    """
    if resource_type not in VALID_RESOURCE_NAMES_IN_ID:
        return False

    prefix = f"{resource_type}-"
    if not full_id_str.startswith(prefix):
        return False

    id_str = full_id_str[len(prefix) :]

    if resource_type == "task":
        return bool(DEADLINE_TASK_ID_PATTERN.fullmatch(id_str))
    else:
        if len(id_str) != 32:
            return False
        return bool(DEADLINE_ID_PATTERN.fullmatch(id_str))


def install_deadline_web_url_handler(all_users: bool) -> None:
    """
    Installs the called AWS Deadline Cloud CLI command as the deadline:// web URL handler.
    """
    if sys.platform == "win32":
        import winreg

        # Get the CLI program path, either an .exe or a .py with the Python interpreter
        deadline_cli_program = os.path.abspath(sys.argv[0])
        if deadline_cli_program.endswith(".py"):
            deadline_cli_prefix = f'"{sys.executable}" "{deadline_cli_program}"'
        else:
            deadline_cli_program = deadline_cli_program + ".exe"
            deadline_cli_prefix = f'"{deadline_cli_program}"'

        if not os.path.isfile(deadline_cli_program):
            raise DeadlineOperationError(
                f"Error determining the AWS Deadline Cloud CLI program, {deadline_cli_program} does not exist."
            )

        logger.info(
            f'Installing "{deadline_cli_program}" as the handler for {DEADLINE_URL_SCHEME_NAME} URLs'
        )

        try:
            hkey: Any = None
            hkey_command: Any = None
            if all_users:
                hkey = winreg.CreateKeyEx(winreg.HKEY_CLASSES_ROOT, DEADLINE_URL_SCHEME_NAME)
            else:
                hkey = winreg.CreateKeyEx(
                    winreg.HKEY_CURRENT_USER, f"Software\\Classes\\{DEADLINE_URL_SCHEME_NAME}"
                )
            winreg.SetValueEx(hkey, None, 0, winreg.REG_SZ, "URL:AWS Deadline Cloud Protocol")
            winreg.SetValueEx(hkey, "URL Protocol", 0, winreg.REG_SZ, "")
            hkey_command = winreg.CreateKeyEx(hkey, "shell\\open\\command")
            winreg.SetValueEx(
                hkey_command,
                None,
                0,
                winreg.REG_SZ,
                f'{deadline_cli_prefix} handle-web-url "%1" --prompt-when-complete',
            )
        except OSError as e:
            if all_users and e.winerror == 5:  # Access denied
                raise DeadlineOperationError(
                    f"Administrator access is required to install the {DEADLINE_URL_SCHEME_NAME} URL handler for all users:\n{e}"
                )
            else:
                raise DeadlineOperationError(
                    f"Failed to install the handler for {DEADLINE_URL_SCHEME_NAME} URLs:\n{e}"
                )
        finally:
            if hkey_command:
                winreg.CloseKey(hkey_command)
            if hkey:
                winreg.CloseKey(hkey)

    elif sys.platform == "linux":
        import subprocess
        import shutil

        if shutil.which("update-desktop-database") is None:
            raise DeadlineOperationError(
                f"Failed to install the handler for {DEADLINE_URL_SCHEME_NAME} URLs: update-desktop-database is not installed."
            )

        # Get the CLI program path
        deadline_cli_program = os.path.abspath(sys.argv[0])

        if all_users:
            entry_dir = "/usr/share/applications"
            mimeapps_list_file_path = "/usr/share/applications/mimeapps.list"
        else:
            entry_dir = os.path.expanduser("~/.local/share/applications")
            mimeapps_list_file_path = os.path.expanduser("~/.config/mimeapps.list")
        try:
            os.makedirs(entry_dir, exist_ok=True)
        except OSError as e:
            raise DeadlineOperationError(f"Failed to create a directory: {e}")

        desktop_file_path = os.path.join(entry_dir, "deadline.desktop")

        desktop_file_content = f"""[Desktop Entry]
Type=Application
Name={DEADLINE_URL_SCHEME_NAME}
Exec={deadline_cli_program} handle-web-url %u
Type=Application
Terminal=true
MimeType=x-scheme-handler/{DEADLINE_URL_SCHEME_NAME}
"""

        mimeapps_file_content = f"""[Default Applications]
x-scheme-handler/{DEADLINE_URL_SCHEME_NAME}={DEADLINE_URL_SCHEME_NAME}.desktop;
"""
        with open(desktop_file_path, "w") as desktop_file:
            desktop_file.write(desktop_file_content)
        with open(mimeapps_list_file_path, "w") as mimeapps_list_file:
            mimeapps_list_file.write(mimeapps_file_content)

        try:
            subprocess.run(["update-desktop-database", entry_dir], check=True)
        except subprocess.CalledProcessError as e:
            raise DeadlineOperationError(
                f"Failed to install the handler for {DEADLINE_URL_SCHEME_NAME} URLs:\n{e}"
            ) from e

    else:
        raise DeadlineOperationError(
            f"Installing the web URL handler is not supported on OS {sys.platform}"
        )


def uninstall_deadline_web_url_handler(all_users: bool) -> None:
    """
    Uninstalls the called AWS Deadline Cloud CLI command as the deadline:// web URL handler.
    """
    if sys.platform == "win32":
        import winreg

        logger.info(f"Unstalling the handler for {DEADLINE_URL_SCHEME_NAME} URLs")

        try:
            hkey: Any = None
            if all_users:
                hkey = winreg.HKEY_CLASSES_ROOT
            else:
                hkey = winreg.OpenKeyEx(winreg.HKEY_CURRENT_USER, "Software\\Classes")
            winreg.DeleteKeyEx(hkey, f"{DEADLINE_URL_SCHEME_NAME}\\shell\\open\\command")
            winreg.DeleteKeyEx(hkey, f"{DEADLINE_URL_SCHEME_NAME}\\shell\\open")
            winreg.DeleteKeyEx(hkey, f"{DEADLINE_URL_SCHEME_NAME}\\shell")
            winreg.DeleteKeyEx(hkey, DEADLINE_URL_SCHEME_NAME)
        except OSError as e:
            if e.winerror == 2:  # Cannot find the specified key
                click.echo(
                    f"Nothing to uninstall, no handler for {DEADLINE_URL_SCHEME_NAME} URLs was installed"
                )
            else:
                raise DeadlineOperationError(
                    f"Failed to uninstall handler for {DEADLINE_URL_SCHEME_NAME} URLs:\n{e}"
                )
        finally:
            if hkey and hkey != winreg.HKEY_CLASSES_ROOT:
                winreg.CloseKey(hkey)

    elif sys.platform == "linux":
        import subprocess
        import shutil

        if shutil.which("update-desktop-database") is None:
            raise DeadlineOperationError(
                f"Failed to uninstall the handler for {DEADLINE_URL_SCHEME_NAME} URLs: update-desktop-database is not installed."
            )

        logger.info(f"Unstalling the handler for {DEADLINE_URL_SCHEME_NAME} URLs")

        if all_users:
            entry_dir = "/usr/share/applications"
        else:
            entry_dir = os.path.expanduser("~/.local/share/applications")
        desktop_file_path = f"{entry_dir}/{DEADLINE_URL_SCHEME_NAME}.desktop"

        try:
            os.remove(desktop_file_path)
            print(f"Removed {desktop_file_path}")
        except FileNotFoundError:
            print(f"{desktop_file_path} not found, nothing to remove")

        try:
            subprocess.run(["update-desktop-database", entry_dir], check=True)
        except subprocess.CalledProcessError as e:
            raise DeadlineOperationError(
                f"Failed to uninstall the handler for {DEADLINE_URL_SCHEME_NAME} URLs:\n{e}"
            ) from e

    else:
        raise DeadlineOperationError(
            f"Uninstalling the web URL handler is not supported on OS {sys.platform}"
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations

__all__ = ["_list_jobs_by_filter_expression"]

from typing import Any
import boto3

from deadline.client.api._session import get_session_client
from botocore.exceptions import ClientError
from deadline.client.exceptions import DeadlineOperationError


class JobFetchFailure(RuntimeError):
    """
    Failure fetching jobs updated since timestamp
    """


def _list_jobs_by_filter_expression(
    boto3_session: boto3.Session,
    farm_id: str,
    queue_id: str,
    filter_expression: dict[str, Any],
) -> list[dict[str, Any]]:
    """
    This function retrieves all jobs in the queue that satisfy a provided filter expression, except potentially
    some jobs updated recently due to eventual consistency. The value is the same as the boto3 filterExpression to
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/deadline/client/search_jobs.html,
    except it may not contain a nested groupFilter.

    CAUTION:
        Eventual consistency in the deadline:SearchJobs API means that the result set can be missing jobs
        with a timestamp close to the current time.

    TODO: This is an experimental function under development, and is exposed under the internal-only name
          deadline.client.api._list_jobs_by_filter_expression._list_jobs_by_filter_expression.
          If it proves useful, deadline.client.api.list_jobs_active is the likely public function.

    NOTE:
        There's an edge case where 100 jobs with identical createdAt timestamps will cause the function to raise
        a JobFetchFailure exception. Because this would require all 100 jobs be created/updated with
        identical timestamp recorded at millisecond precision, we do not expect this to occur in practice.

    Example:
        boto3_session = boto3.Session()
        farm_id = ...
        queue_id = ...
        saved_timestamp = ...

        # Get the jobs created in the last 5 minutes
        all_active_jobs = _list_jobs_by_filter_expression(
            boto3_session,
            farm_id,
            queue_id,
        )

    Args:
      boto3_session (boto3.Session): The boto3 Session for AWS API access.
      farm_id (str): The Farm ID.
      queue_id (str): The Queue ID.
      filter_expressions (dict[str, Any]): The filter expression to apply to jobs. This is nested one level in a
            filter expression provided to deadline:SearchJobs, so cannot include a groupFilter.

    Returns:
      The list of all jobs in the queue that satisfy the provided filter expression. Each job is as returned by the deadline:SearchJobs API.
    """
    # This function uses deadline:SearchJobs as a primitive to union subsets of up to 100 jobs at a time,
    # designed to guarantee we end up with the full correct set. Ordering and thresholding by the createdAt
    # timestamp ensures no jobs get missed as might occur if we incremented itemOffset instead.
    #
    # Let J = {j ∈ queue | filter_expression(j)} be the set we want
    #     J' = union {J_i | i ∈ range(search_count)} be the set the algorithm produces,
    #            where J_0 = {j ∈ queue | filter_expression(j)} [limit 100, ordered by createdAt asc]
    #                  J_i = {j ∈ queue | filter_expression(j) && j["createdAt"] >= timestamp_i} [when i > 0, limit 100, ordered by createdAt asc]
    #                  timestamp_i = max {j["timestamp_field_name"]: j in J_{i-1}} [when i >= 1]
    #
    # To prove these sets are equal, we show they are each a subset of the other.
    #
    # J' is a subset of J, because each J_i is from a deadline:SearchJobs that filters to a subset of J.
    #
    # Consider a job j ∈ J. The createdAt timestamp of a job is always available, and never changes. By construction,
    # the set J_0 is equivalent to {j ∈ queue | filter_expression(j) && j["createdAt"] <= timestamp_1}, and each
    # subsequent J_i is equivalent to {j ∈ queue | filter_expression(j) && timestamp_i <= j["createdAt"] <= timestamp_[i+1]},
    # except for when there are more than 100 jobs with identical createdAt timestamp, in which case the algorithm
    # raises an exception. All of the createdAt timestamps in J are within the range of timestamps covered by these sets,
    # so therefore J is also a subset of J'.
    #
    # The value of filter_expression(j) may change between subsequent calls to deadline:SearchJob. If it does, the job
    # is within the eventual consistency window.

    # Do some basic parameter validation on filter_expression. The rest is left up to the deadline:SearchJobs API handler.
    if not isinstance(filter_expression, dict):
        raise ValueError("The provided filter expression must be a dict")

    if sorted(filter_expression.keys()) != ["filters", "operator"]:
        raise ValueError(
            f"The provided filter expression must contain 'filters' and 'operator', got {sorted(filter_expression.keys())}"
        )

    # This holds {job_id: job_from_search_jobs_call, ...}
    result_jobs = {}

    deadline = get_session_client(boto3_session, "deadline")

    # Sort jobs in ascending order of the timestamp field
    sort_expressions = [{"fieldSort": {"name": "CREATED_AT", "sortOrder": "ASCENDING"}}]

    # Filter for any of the active statuses READY, ASSIGNED, STARTING, SCHEDULED, or RUNNING
    provided_filter = {
        "groupFilter": filter_expression,
    }

    # The first time we call deadline.search_jobs, there is no timestamp filter so it
    # will return the earliest jobs satisfying the filter, ordered by createdAt.
    query_filter_expressions = {
        "filters": [provided_filter],
        "operator": "AND",
    }

    # Continue until we've processed all jobs
    while True:
        try:
            # The pageSize defaults to its maximum, 100, so we leave it out of the call.
            response = deadline.search_jobs(
                farmId=farm_id,
                queueIds=[queue_id],
                itemOffset=0,
                filterExpressions=query_filter_expressions,
                sortExpressions=sort_expressions,
            )
        except ClientError as exc:
            raise DeadlineOperationError(f"Failed to get Jobs from Deadline:\n{exc}") from exc

        # This is up to the first 100 of jobs that satisfy the query
        jobs = response.get("jobs", [])
        # This is the total number of jobs that satisfied the query
        total_results = response.get("totalResults", 0)

        result_jobs.update({job["jobId"]: job for job in jobs})

        if len(jobs) == total_results:
            # If the jobs we got are the total results, result_jobs is now the full set
            break
        elif jobs[0]["createdAt"] == jobs[-1]["createdAt"]:
            # Rare edge case where the timestamp field is the same for all 100 jobs in the page, that
            # we expect to never see in practice. The timestamp value is stored with
            # millisecond precision, and jobs are scheduled independently from each other,
            # with updates of running jobs generally being multiple seconds apart.
            raise JobFetchFailure(
                "Failure fetching jobs based on the createdAt field as more then 100 jobs have the exact same timestamp value."
            )
        else:
            # Continue processing from the largest timestamp value we saw so far
            threshold_timestamp = jobs[-1]["createdAt"]

            # Update jobs to have job["createdAt"] >= threshold_timestamp
            query_filter_expressions = {
                "filters": [
                    provided_filter,
                    {
                        "dateTimeFilter": {
                            "name": "CREATED_AT",
                            "dateTime": threshold_timestamp,
                            "operator": "GREATER_THAN_EQUAL_TO",
                        }
                    },
                ],
                "operator": "AND",
            }

    return list(result_jobs.values())



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline worker` commands.
"""

import click
from botocore.exceptions import ClientError  # type: ignore[import]

from ... import api
from ...config import config_file
from ...exceptions import DeadlineOperationError
from .._common import _apply_cli_options_to_config, _cli_object_repr, _handle_error
from .._main import main


@main.group(name="worker")
@_handle_error
def cli_worker():
    """
    Commands to work with workers.
    """


@cli_worker.command(name="list")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--fleet-id", help="The fleet to use.", required=True)
@click.option("--page-size", default=5, help="The number of workers to load at a time.")
@click.option("--item-offset", default=0, help="The index of the worker to start listing from.")
@_handle_error
def worker_list(page_size, item_offset, fleet_id, **args):
    """
    Lists the Workers in a fleet.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    try:
        response = deadline.search_workers(
            farmId=farm_id, fleetIds=[fleet_id], itemOffset=item_offset, pageSize=page_size
        )
    except ClientError as exc:
        raise DeadlineOperationError(f"Failed to get Workers from Deadline:\n{exc}") from exc

    total_results = response["totalResults"]

    # Select which fields to print and in which order
    structured_worker_list = [
        {field: worker[field] for field in ["workerId", "status", "createdAt"]}
        for worker in response["workers"]
    ]

    click.echo(
        f"Displaying {len(structured_worker_list)} of {total_results} workers starting at {item_offset}"
    )
    click.echo()
    click.echo(_cli_object_repr(structured_worker_list))


@cli_worker.command(name="get")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--fleet-id", help="The fleet to use.", required=True)
@click.option("--worker-id", help="The worker to get.", required=True)
@_handle_error
def worker_get(fleet_id, worker_id, **args):
    """
    Get the details of a worker.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    response = deadline.get_worker(farmId=farm_id, fleetId=fleet_id, workerId=worker_id)
    response.pop("ResponseMetadata", None)

    click.echo(_cli_object_repr(response))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
from pathlib import Path
import shutil
import sys
from dataclasses import dataclass
from enum import Enum
from typing import List, Set, Union

from .exceptions import AssetSyncError, PathOutsideDirectoryError
from ._utils import _is_relative_to, _normalize_windows_path


@dataclass
class PosixFileSystemPermissionSettings:
    """
    A dataclass representing file system permission-related information
    for Posix. The specified permission modes will be bitwise-OR'ed with
    the directory or file's existing permissions.

    Attributes:
        os_user (str): The target operating system user for ownership.
        os_group (str): The target operating system group for ownership.
        dir_mode (int): The permission mode to be added to directories.
        file_mode (int): The permission mode to be added to files.
    """

    os_user: str
    os_group: str
    dir_mode: int
    file_mode: int


@dataclass
class WindowsPermissionEnum(Enum):
    """
    An enumeration of different Windows permission flags.
    """

    READ = "READ"
    WRITE = "WRITE"
    EXECUTE = "EXECUTE"
    READ_WRITE = "READ_WRITE"
    FULL_CONTROL = "FULL_CONTROL"


@dataclass
class WindowsFileSystemPermissionSettings:
    """
    A dataclass representing file system permission-related information
    for Windows.

    Attributes:
        os_user (str): The target operating system user or ownership.
        os_group (str): The target operating system group for ownership.
        dir_mode (WindowsPermissionEnum): The permission mode to be added to directories.
        file_mode (WindowsPermissionEnum): The permission mode to be added to files.
    """

    os_user: str
    dir_mode: WindowsPermissionEnum
    file_mode: WindowsPermissionEnum


# A union of different file system permission settings that are based on the underlying OS.
FileSystemPermissionSettings = Union[
    PosixFileSystemPermissionSettings, WindowsFileSystemPermissionSettings
]


def _set_fs_group_for_posix(
    file_paths: List[str],
    local_root: str,
    fs_permission_settings: PosixFileSystemPermissionSettings,
) -> None:
    os_group = fs_permission_settings.os_group
    dir_mode = fs_permission_settings.dir_mode
    file_mode = fs_permission_settings.file_mode

    # A set that stores the unique directory paths where permissions need to be changed.
    dir_paths_to_change_fs_group: Set[Path] = set()

    # 1. Set group ownership and permissions for each file.
    for file_path_str in file_paths:
        # The file path must be relative to the root path (ie. local_root).
        if not _is_relative_to(file_path_str, local_root):
            raise PathOutsideDirectoryError(
                f"The provided path '{file_path_str}' is not under the root directory: {local_root}"
            )

        _change_permission_for_posix(file_path_str, os_group, file_mode)

        # Add the parent directories of each file to the set of directories whose
        # group ownership and permissions will be changed.
        path_components = Path(file_path_str).relative_to(local_root).parents
        for path_component in path_components:
            path_to_change = Path(local_root).joinpath(path_component)
            dir_paths_to_change_fs_group.add(path_to_change)

    # 2. Set group ownership and permissions for the directories in the path starting from root.
    for dir_path in dir_paths_to_change_fs_group:
        _change_permission_for_posix(str(dir_path), os_group, dir_mode)


def _set_fs_permission_for_windows(
    file_paths: List[str],
    local_root: str,
    fs_permission_settings: WindowsFileSystemPermissionSettings,
) -> None:
    os_user = fs_permission_settings.os_user
    dir_mode = fs_permission_settings.dir_mode
    file_mode = fs_permission_settings.file_mode

    # A set that stores the unique directory paths where permissions need to be changed.
    dir_paths_to_change_fs_group: Set[Path] = set()

    # 1. Set permissions for each file.
    for file_path_str in file_paths:
        # The file path must be relative to the root path (ie. local_root).
        if not _is_relative_to(file_path_str, local_root):
            raise PathOutsideDirectoryError(
                f"The provided path '{file_path_str}' is not under the root directory: {local_root}"
            )

        _change_permission_for_windows(file_path_str, os_user, file_mode)

        # Add the parent directories of each file to the set of directories whose
        # permissions will be changed.
        path_components = (
            _normalize_windows_path(file_path_str)
            .relative_to(_normalize_windows_path(local_root))
            .parents
        )
        for path_component in path_components:
            path_to_change = Path(local_root).joinpath(path_component)
            dir_paths_to_change_fs_group.add(path_to_change)

    # 2. Set permissions for the directories in the path starting from root.
    for dir_path in dir_paths_to_change_fs_group:
        _change_permission_for_windows(str(dir_path), os_user, dir_mode)


def _change_permission_for_posix(
    path_str: str,
    os_group: str,
    mode: int,
) -> None:
    if sys.platform == "win32":
        raise EnvironmentError("This function can only be executed on POSIX systems.")

    path = Path(path_str)
    shutil.chown(path, group=os_group)
    os.chmod(path, path.stat().st_mode | mode)


def _change_permission_for_windows(
    path: str,
    os_user: str,
    mode: WindowsPermissionEnum,
) -> None:
    if sys.platform != "win32":
        raise EnvironmentError("This function can only be executed on Windows systems.")

    import win32security

    try:
        con_mode = _get_ntsecuritycon_mode(mode)
        # Lookup the user's SID (Security Identifier)
        user_sid = win32security.LookupAccountName(None, os_user)[0]
        # Get existing DACL (Discretionary Access Control List). If dacl is none, create a new one.
        sd = win32security.GetFileSecurity(path, win32security.DACL_SECURITY_INFORMATION)
        dacl = sd.GetSecurityDescriptorDacl()
        if dacl is None:
            dacl = win32security.ACL()
        # Add new ACE (Access Control Entry)
        dacl.AddAccessAllowedAce(win32security.ACL_REVISION, con_mode, user_sid)
        # Set the modified DACL to the security descriptor
        sd.SetSecurityDescriptorDacl(1, dacl, 0)
        win32security.SetFileSecurity(path, win32security.DACL_SECURITY_INFORMATION, sd)
    except win32security.error as e:
        raise AssetSyncError(
            f"Failed to set permissions for file or directory ({path}): {e}"
        ) from e


def _get_ntsecuritycon_mode(mode: WindowsPermissionEnum) -> int:
    """
    Get the NTSecurityCon mode for a WindowsPermissionEnum.
    """
    if sys.platform != "win32":
        raise EnvironmentError("This function can only be executed on Windows systems.")

    import ntsecuritycon as con

    permission_mapping = {
        WindowsPermissionEnum.READ.value: con.FILE_GENERIC_READ,
        WindowsPermissionEnum.WRITE.value: con.FILE_GENERIC_WRITE,
        WindowsPermissionEnum.EXECUTE.value: con.FILE_GENERIC_EXECUTE,
        WindowsPermissionEnum.READ_WRITE.value: con.FILE_GENERIC_READ | con.FILE_GENERIC_WRITE,
        WindowsPermissionEnum.FULL_CONTROL.value: con.FILE_ALL_ACCESS,
    }
    return permission_mapping[mode.value]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
UI widgets for the job attachments tab.
"""

from __future__ import annotations
import os
from logging import getLogger
from typing import Optional

from qtpy.QtWidgets import (  # type: ignore
    QAbstractItemView,
    QFileDialog,
    QHBoxLayout,
    QCheckBox,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QSizePolicy,
    QLabel,
    QVBoxLayout,
    QWidget,
    QGroupBox,
    QMessageBox,
)

from ...job_bundle.submission import AssetReferences
from .. import block_signals

logger = getLogger(__name__)


class JobAttachmentsWidget(QWidget):
    """
    The Widget for showing which files and folders will be attached, and to let the user configure more.
    The files and folders in auto_detected_attachments cannot be removed from the list, but they can
    be hidden or shown.

    Args:
        auto_detected_attachments (FlatAssetReferences): The job attachments that were automatically detected
            from the input document/scene file or starting job bundle.
        attachments: (FlatAssetReferences): The job attachments that have been added to the job by the user.
    """

    def __init__(
        self,
        auto_detected_attachments: AssetReferences,
        attachments: AssetReferences,
        parent: QWidget = None,
    ) -> None:
        super().__init__(parent=parent)
        self.auto_detected_attachments = auto_detected_attachments
        self.attachments = attachments

        self._build_ui()
        self._populate_attachment_lists()

    def _build_ui(self) -> None:
        tab_layout = QVBoxLayout(self)

        # Create a group box for general settings
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        size_policy.setVerticalStretch(10)
        self.general_group = QGroupBox("General submission settings", self)
        tab_layout.addWidget(self.general_group)
        self.general_group.setSizePolicy(size_policy)
        general_layout = QVBoxLayout(self.general_group)

        # Create a group box for each type of attachment
        self.input_files_group = QGroupBox("Attach input files", self)
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        size_policy.setVerticalStretch(80)
        tab_layout.addWidget(self.input_files_group)
        self.input_files_group.setSizePolicy(size_policy)
        input_files_layout = QVBoxLayout(self.input_files_group)

        self.input_directories_group = QGroupBox("Attach input directories", self)
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        size_policy.setVerticalStretch(10)
        self.input_directories_group.setSizePolicy(size_policy)
        tab_layout.addWidget(self.input_directories_group)
        input_directories_layout = QVBoxLayout(self.input_directories_group)

        self.output_directories_group = QGroupBox("Specify output directories", self)
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        size_policy.setVerticalStretch(10)
        self.output_directories_group.setSizePolicy(size_policy)
        tab_layout.addWidget(self.output_directories_group)
        output_directories_layout = QVBoxLayout(self.output_directories_group)

        # General settings
        self.general_settings = JobAttachmentsGeneralWidget(self)
        general_layout.addWidget(self.general_settings)

        # The "Attach Input Files" attachments
        self.input_files_controls = JobAttachmentsControlsWidget(self)
        self.input_files_controls.show_auto_detected.stateChanged.connect(
            self.show_auto_detected_change
        )
        self.input_files_controls.add.clicked.connect(self._add_input_files)
        self.input_files_controls.remove_selected.clicked.connect(self._remove_selected_input_files)
        self.input_files = QListWidget(parent=self)
        self.input_files.itemSelectionChanged.connect(self._update_status_messages)
        self.input_files.setSortingEnabled(False)
        self.input_files.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.input_files.setAlternatingRowColors(True)

        input_files_layout.addWidget(self.input_files_controls)
        input_files_layout.addWidget(self.input_files)

        # The "Attach input directories" attachments
        self.input_directories_controls = JobAttachmentsControlsWidget(self)
        self.input_directories_controls.show_auto_detected.stateChanged.connect(
            self.show_auto_detected_change
        )
        self.input_directories_controls.add.clicked.connect(self._add_input_directory)
        self.input_directories_controls.remove_selected.clicked.connect(
            self._remove_selected_input_directories
        )
        self.input_directories = QListWidget(parent=self)
        self.input_directories.itemSelectionChanged.connect(self._update_status_messages)
        self.input_directories.setSortingEnabled(False)
        self.input_directories.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.input_directories.setAlternatingRowColors(True)

        input_directories_layout.addWidget(self.input_directories_controls)
        input_directories_layout.addWidget(self.input_directories)

        # The "Specify output directories" attachments
        self.output_directories_controls = JobAttachmentsControlsWidget(self)
        self.output_directories_controls.show_auto_detected.stateChanged.connect(
            self.show_auto_detected_change
        )
        self.output_directories_controls.add.clicked.connect(self._add_output_directory)
        self.output_directories_controls.remove_selected.clicked.connect(
            self._remove_selected_output_directories
        )
        self.output_directories = QListWidget(parent=self)
        self.output_directories.itemSelectionChanged.connect(self._update_status_messages)
        self.output_directories.setSortingEnabled(False)
        self.output_directories.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.output_directories.setAlternatingRowColors(True)

        output_directories_layout.addWidget(self.output_directories_controls)
        output_directories_layout.addWidget(self.output_directories)

        self.attachments_controls = self._build_attachment_controls()

    def _build_attachment_controls(
        self,
    ) -> list[tuple[set[str], set[str], QListWidget, JobAttachmentsControlsWidget]]:
        # Put all the controls in a list of tuples for structured processing
        return [
            (
                self.auto_detected_attachments.input_filenames,
                self.attachments.input_filenames,
                self.input_files,
                self.input_files_controls,
            ),
            (
                self.auto_detected_attachments.input_directories,
                self.attachments.input_directories,
                self.input_directories,
                self.input_directories_controls,
            ),
            (
                self.auto_detected_attachments.output_directories,
                self.attachments.output_directories,
                self.output_directories,
                self.output_directories_controls,
            ),
        ]

    def refresh_ui(
        self,
        auto_detected_attachments: Optional[AssetReferences],
        attachments: Optional[AssetReferences],
    ):
        """Refresh the job attachment lists if provided"""
        if auto_detected_attachments:
            self.auto_detected_attachments = auto_detected_attachments
        if attachments:
            self.attachments = attachments
        self.attachments_controls = self._build_attachment_controls()
        self._populate_attachment_lists()

    def _set_attachments_list(
        self, list_widget: QListWidget, auto_detected_paths: set[str], paths: set[str]
    ):
        with block_signals(list_widget):
            list_widget.clear()
            if auto_detected_paths:
                italic_font = list_widget.font()
                italic_font.setItalic(True)
                for path in auto_detected_paths:
                    item = QListWidgetItem()
                    item.setText(path)
                    item.setFont(italic_font)
                    list_widget.addItem(item)
            list_widget.addItems(list(paths))
            list_widget.sortItems()

    def _populate_attachment_lists(self) -> None:
        """Initializes the lists of job attachments. Removes any auto-detected attachments from the added attachments sets."""
        for auto_detected_set, added_set, list_widget, controls_widget in self.attachments_controls:
            added_set.difference_update(auto_detected_set)
            if controls_widget.show_auto_detected.isChecked():
                self._set_attachments_list(
                    list_widget,
                    auto_detected_set,
                    added_set,
                )
            else:
                self._set_attachments_list(
                    list_widget,
                    set(),
                    added_set,
                )
        self._update_status_messages()

    def show_auto_detected_change(self, _: int):
        # When one of the "Show Auto-detected" checkboxes changes state, re-populate the lists
        self._populate_attachment_lists()

    def _update_status_messages(self) -> None:
        """Updates the status messages for each list to have the counts."""
        for auto_detected_set, added_set, list_widget, controls_widget in self.attachments_controls:
            selected_count = len(list_widget.selectedItems())
            controls_widget.status_message.setText(
                f"{len(auto_detected_set)} auto, {len(added_set)} added, {selected_count} selected"
            )

    def _add_input_files(self) -> None:
        new_files, _ = QFileDialog.getOpenFileNames(
            self, "Select input files to attach to your job"
        )

        if new_files:
            # Normalize the paths
            paths = set(os.path.normpath(path) for path in new_files)
            # Remove any that are from the auto-detected set
            paths.difference_update(self.auto_detected_attachments.input_filenames)
            # Add them to the attachments
            self.attachments.input_filenames.update(paths)
            self._populate_attachment_lists()

    def _remove_selected_input_files(self) -> None:
        selected_files = [item.text() for item in self.input_files.selectedItems()]
        unremoved_files = self.auto_detected_attachments.input_filenames.intersection(
            selected_files
        )
        # Remove the selected items only from the added attachments, not the auto-detected ones
        self.attachments.input_filenames.difference_update(selected_files)
        self._populate_attachment_lists()
        if unremoved_files:
            QMessageBox.warning(  # type: ignore[call-arg]
                self,
                "Some files were not removed",
                f"The selected files from the auto-detected list ({len(unremoved_files)} items) were not removed.",
            )

    def _add_input_directory(self) -> None:
        new_dir = QFileDialog.getExistingDirectory(self, "Select a directory to attach to your job")

        if new_dir:
            # Normalize the path
            new_dir = os.path.normpath(new_dir)
            # Check if it's in the auto-detected set
            if new_dir not in self.auto_detected_attachments.input_directories:
                # Add it to the attachments
                self.attachments.input_directories.add(new_dir)
                self._populate_attachment_lists()

    def _remove_selected_input_directories(self) -> None:
        selected_dirs = [item.text() for item in self.input_directories.selectedItems()]
        unremoved_dirs = self.auto_detected_attachments.input_directories.intersection(
            selected_dirs
        )
        # Remove the selected items only from the added attachments, not the auto-detected ones
        self.attachments.input_directories.difference_update(selected_dirs)
        self._populate_attachment_lists()
        if unremoved_dirs:
            QMessageBox.warning(  # type: ignore[call-arg]
                self,
                "Some directories were not removed",
                f"The selected directories from the auto-detected list ({len(unremoved_dirs)} items) were not removed.",
            )

    def _add_output_directory(self) -> None:
        new_dir = QFileDialog.getExistingDirectory(self, "Select an output directory of your job")

        if new_dir:
            # Normalize the path
            new_dir = os.path.normpath(new_dir)
            # Check if it's in the auto-detected set
            if new_dir not in self.auto_detected_attachments.output_directories:
                # Add it to the attachments
                self.attachments.output_directories.add(new_dir)
                self._populate_attachment_lists()

    def _remove_selected_output_directories(self) -> None:
        selected_dirs = [item.text() for item in self.output_directories.selectedItems()]
        unremoved_dirs = self.auto_detected_attachments.output_directories.intersection(
            selected_dirs
        )
        # Remove the selected items only from the added attachments, not the auto-detected ones
        self.attachments.output_directories.difference_update(selected_dirs)
        self._populate_attachment_lists()
        if unremoved_dirs:
            QMessageBox.warning(  # type: ignore[call-arg]
                self,
                "Some directories were not removed",
                f"The selected directories from the auto-detected list ({len(unremoved_dirs)} items) were not removed.",
            )

    def get_asset_references(self) -> AssetReferences:
        """
        Creates an asset_references object that can be saved as the
        asset_references.json|yaml file in a Job Bundle.
        """
        return self.auto_detected_attachments.union(self.attachments)

    def get_require_paths_exist(self) -> bool:
        """
        Returns the checkbox value of whether to allow empty paths or not.
        """
        return self.general_settings.require_paths_exist.isChecked()


class JobAttachmentsGeneralWidget(QWidget):
    """
    A Widget that contains general settings for a specific submission.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.require_paths_exist = QCheckBox("Require all input paths exist", parent=self)
        layout.addWidget(self.require_paths_exist)


class JobAttachmentsControlsWidget(QWidget):
    """
    A Widget that contains buttons for the actions to add/remove files or directories
    in a job attachments list.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.add = QPushButton("Add...", parent=self)
        layout.addWidget(self.add)

        self.remove_selected = QPushButton("Remove selected", parent=self)
        layout.addWidget(self.remove_selected)

        self.status_message = QLabel("0 total")
        self.status_message.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        layout.addWidget(self.status_message)

        self.show_auto_detected = QCheckBox("Show auto-detected", parent=self)
        self.show_auto_detected.setChecked(True)
        layout.addWidget(self.show_auto_detected)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline config` commands.
"""

import click
import json
import textwrap

from ...config import config_file
from .._common import _handle_error
from .._main import main


@main.group(name="config")
@_handle_error
def cli_config():
    """
    Manage Deadline's workstation configuration.
    """


@cli_config.command(name="show")
@click.option(
    "--output",
    type=click.Choice(["verbose", "json"], case_sensitive=False),
    default="verbose",
    help="Output format of the command",
)
@_handle_error
def config_show(output):
    """
    Show all workstation configuration settings and current values.
    """
    settings_json = {}
    if output == "verbose":
        click.echo(
            f"AWS Deadline Cloud configuration file:\n   {config_file.get_config_file_path()}"
        )
        click.echo()
        for setting_name in config_file.SETTINGS.keys():
            setting_value = config_file.get_setting(setting_name)
            setting_default = config_file.get_setting_default(setting_name)

            # Wrap and indent the descriptions to 80 characters because they may be multiline.
            setting_description: str = config_file.SETTINGS[setting_name].get("description", "")
            setting_description = "\n".join(
                f"   {line}" for line in textwrap.wrap(setting_description, width=77)
            )
            click.echo(
                f"{setting_name}: {setting_value} {'(default)' if setting_value == setting_default else ''}"
            )
            click.echo(setting_description)
            click.echo()
    else:
        settings_json["settings.config_file_path"] = str(config_file.get_config_file_path())
        for setting_name in config_file.SETTINGS.keys():
            setting_value = config_file.get_setting(setting_name)
            settings_json[setting_name] = setting_value

        click.echo(json.dumps(settings_json))


@cli_config.command(name="gui")
@click.option(
    "--install-gui",
    is_flag=True,
    help="Installs GUI dependencies if they are not installed already",
)
@_handle_error
def config_gui(install_gui: bool):
    """
    Open the workstation configuration settings GUI.
    """
    from ...ui import gui_context_for_cli

    with gui_context_for_cli(automatically_install_dependencies=install_gui):
        from ...ui.dialogs.deadline_config_dialog import DeadlineConfigDialog

        DeadlineConfigDialog.configure_settings()


@cli_config.command(name="set")
@click.argument("setting_name")
@click.argument("value")
@_handle_error
def config_set(setting_name, value):
    """
    Sets a workstation configuration setting.

    For example `deadline config set defaults.farm_id <farm-id>`.
    Run `deadline config --help` to show available settings.
    """
    config_file.set_setting(setting_name, value)


@cli_config.command(name="clear")
@click.argument("setting_name")
@_handle_error
def config_clear(setting_name):
    """
    Sets a workstation configuration setting back to the default value.

    For example `deadline config clear defaults.farm_id`.
    Run `deadline config --help` to show available settings.
    """
    config_file.clear_setting(setting_name)


@cli_config.command(name="get")
@click.argument("setting_name")
@_handle_error
def config_get(setting_name):
    """
    Gets a workstation configuration setting.

    For example `deadline config get defaults.farm_id`.
    Run `deadline config --help` to show available settings.
    """
    click.echo(config_file.get_setting(setting_name))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from enum import Enum

# ---- Types to export


class JobBundlePurpose(str, Enum):
    EXPORT = "export"
    """A job bundle is being created for export."""

    SUBMISSION = "submission"
    """A job bundle is being created for immediate submission."""



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import signal


class SigIntHandler:
    """
    A singleton class to handle SIGINT signals (triggered by Ctrl + C). Sets
    a flag `continue_operation` to False to indicate the interruption of an
    ongoing process. (e.g. job submission, output download)
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
            cls._instance.continue_operation = True
            signal.signal(signal.SIGINT, cls._instance._handle_sigint)
        return cls._instance

    def _handle_sigint(self, signum, frame):
        self.continue_operation = False



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides the function to submit a job bundle to AWS Deadline Cloud.
"""

from __future__ import annotations

import json
import logging
import time
import os
import re
import textwrap
from configparser import ConfigParser
from typing import Any, Callable, Dict, List, Optional, Tuple, Iterable
from collections.abc import Collection
from pathlib import Path

from botocore.client import BaseClient

from .. import api
from ..exceptions import (
    DeadlineOperationError,
    CreateJobWaiterCanceled,
    DeadlineOperationCanceled,
    UserInitiatedCancel,
)
from ..config import get_setting, set_setting, config_file
from ..job_bundle import deadline_yaml_dump
from ..job_bundle.loader import (
    read_yaml_or_json,
    read_yaml_or_json_object,
    parse_yaml_or_json_content,
    validate_directory_symlink_containment,
)
from ..job_bundle.parameters import (
    apply_job_parameters,
    merge_queue_job_parameters,
    read_job_bundle_parameters,
    JobParameter,
)
from ..job_bundle.submission import AssetReferences, split_parameter_args
from ...job_attachments.exceptions import MisconfiguredInputsError
from ...job_attachments.models import (
    JobAttachmentsFileSystem,
    AssetRootManifest,
    AssetUploadGroup,
    JobAttachmentS3Settings,
    StorageProfile,
    FileSystemLocationType,
)
from ...job_attachments.progress_tracker import ProgressReportMetadata, ProgressStatus
from ...job_attachments.upload import S3AssetManager
from ._session import session_context
from ._job_attachment import _hash_attachments  # type: ignore[import]
from ...job_attachments._path_summarization import human_readable_file_size, summarize_path_list

logger = logging.getLogger(__name__)


def _summarize_asset_paths(
    input_paths: Collection[Path | str], output_paths: Collection[Path | str], name: str
) -> list[str]:
    result = []
    if input_paths:
        result.append(f"{name} for upload:\n")
        result.append(textwrap.indent(summarize_path_list(input_paths), "  "))
    if output_paths:
        result.append(f"{name} to collect job outputs for download:\n")
        # We expect the list of output paths to be small, but truncate it to an arbitrary limit just in case for the summary
        summary_entry_count = 4
        if len(output_paths) == summary_entry_count + 1:
            summary_entry_count += 1
        result.extend(f"  {path}\n" for path in sorted(output_paths)[:summary_entry_count])
        if len(output_paths) > summary_entry_count:
            result.append(f"\n  ... and {len(output_paths) - summary_entry_count} more\n")
    return result


def _generate_message_for_asset_paths(
    upload_group: AssetUploadGroup,
    storage_profile: Optional[StorageProfile],
    known_asset_paths: Iterable[str],
) -> tuple[str, bool]:
    """Generate a message about asset uploads and along with a flag indicating if there are warnings."""
    # Collect all the input and output paths
    all_input_paths: set[Path | str] = set()
    all_output_paths: set[Path | str] = set()
    for group in upload_group.asset_groups:
        all_input_paths.update(path for path in group.inputs)
        all_output_paths.update(path for path in group.outputs)

    # Filter to get the unknown paths
    if known_asset_paths:
        known_path_regex = re.compile(
            f"{'|'.join(re.escape(path) for path in known_asset_paths)}.*"
        )
        unknown_input_paths = {
            path for path in all_input_paths if not known_path_regex.match(str(path))
        }
        unknown_output_paths = {
            path for path in all_output_paths if not known_path_regex.match(str(path))
        }
    else:
        unknown_input_paths = all_input_paths
        unknown_output_paths = all_output_paths

    unknown_path_warnings = _summarize_asset_paths(
        unknown_input_paths, unknown_output_paths, "Unknown locations"
    )

    warning_messages = []
    default_prompt_response = not unknown_path_warnings
    if unknown_path_warnings:
        warning_messages.append("\nWARNING: Files were specified outside of known asset paths.\n\n")

    warning_messages.extend(
        [
            f"Job submission contains {upload_group.total_input_files} input files "
            f"totaling {human_readable_file_size(upload_group.total_input_bytes)}. "
            "All input files will be uploaded to S3 if they are not already present in the job attachments bucket.\n\n"
        ]
    )
    warning_messages.extend(_summarize_asset_paths(all_input_paths, all_output_paths, "Locations"))

    if unknown_path_warnings:
        warning_messages.append("\n---\n\n")
        warning_messages.append("The list of known asset prefixes for this submission are:\n")
        if known_asset_paths:
            warning_messages.extend(f"  {path}\n" for path in sorted(set(known_asset_paths)))
        else:
            warning_messages.append("  (empty list)\n")
        warning_messages.append("\n")
        warning_messages.extend(unknown_path_warnings)
        warning_messages.append(
            "\nTo enable submission without user input, add directory locations containing the unknown paths to either "
            + "1. The list of known asset paths in the local Deadline Cloud configuration. "
        )
        if storage_profile:
            warning_messages.append(
                f"2. The Storage Profile '{storage_profile.displayName}' as LOCAL file system locations, from the AWS Deadline Cloud management console.\n"
            )
        else:
            warning_messages.append(
                "2. In a Storage Profile as LOCAL file system locations created from the AWS Deadline Cloud management console, and then configured on your workstation.\n"
            )

    return "".join(warning_messages), default_prompt_response


@api.record_success_fail_telemetry_event(metric_name="asset_upload")
def _upload_attachments(
    asset_manager: S3AssetManager,
    manifests: List[AssetRootManifest],
    print_function_callback: Callable,
    upload_progress_callback: Optional[Callable],
    config: Optional[ConfigParser] = None,
    from_gui: bool = False,
) -> Dict[str, Any]:
    """
    Starts the job attachments upload and handles the progress reporting callback.
    Returns the attachment settings from the upload.
    """

    def _default_update_upload_progress(upload_metadata: ProgressReportMetadata) -> bool:
        return True

    if not upload_progress_callback:
        upload_progress_callback = _default_update_upload_progress

    upload_summary, attachment_settings = asset_manager.upload_assets(
        manifests=manifests,
        on_uploading_assets=upload_progress_callback,
        s3_check_cache_dir=config_file.get_cache_directory(),
    )
    api.get_deadline_cloud_library_telemetry_client(config=config).record_upload_summary(
        upload_summary,
        from_gui=from_gui,
    )

    if upload_summary.total_files > 0:
        print_function_callback("Upload Summary:")
        print_function_callback(textwrap.indent(str(upload_summary), "    "))
    else:
        # Ensure to call the callback once if no files were processed
        upload_progress_callback(
            ProgressReportMetadata(
                status=ProgressStatus.UPLOAD_IN_PROGRESS,
                progress=100,
                transferRate=0,
                progressMessage="No files to upload",
            )
        )

    return attachment_settings.to_dict()


def _filter_redundant_known_paths(known_asset_paths: Iterable[str]) -> list[str]:
    """
    Filters out redundant paths from the known asset paths list.

    This algorithm identifies any paths that have a different path as a prefix,
    and removes them from the list. Pseudo-code is:

        1. Sort the paths from shortest to longest, so any prefix of a path has
           to happen before that path.
        2. For each path, split it into parts (i.e. '/mnt/prod/project' becomes
           ['/', 'mnt', 'prod', 'project']), and then insert it part by part into
           a nested dict called dir_tree organized as a TRIE. The value True in the
           TRIE indicates that a path with that as its final part is in the list.
        3. While inserting a path into the TRIE, detect whether another path already
           had a prefix of the parts, and filter out the path when that occurs.
    """
    # This directory tree gets filled with the known asset paths, with
    # a True value as a marker for the last part of already seen paths.
    dir_tree: dict[str, Any] = {}
    filtered_paths: list[str] = []
    # Process the paths from shortest to longest, so that prefixes are always seen first
    for path in sorted(known_asset_paths, key=len):
        parts = Path(path).parts
        current: Optional[dict[str, Any]] = dir_tree
        for part in parts[:-1]:
            # If we see a True value, another path is a prefix so we can skip it.
            if current.get(part) is True:  # type: ignore
                current = None
                break
            current = current.setdefault(part, {})  # type: ignore
        # If we didn't find a prefix or equal path, add this one and mark it in dir_tree
        if current is not None and current.get(parts[-1]) is not True:
            filtered_paths.append(path)
            current[parts[-1]] = True
    return filtered_paths


@api.record_function_latency_telemetry_event()
def create_job_from_job_bundle(
    job_bundle_dir: str,
    job_parameters: list[dict[str, Any]] = [],
    *,
    name: Optional[str] = None,
    queue_parameter_definitions: Optional[list[JobParameter]] = None,
    job_attachments_file_system: Optional[str] = None,
    config: Optional[ConfigParser] = None,
    priority: Optional[int] = None,
    max_failed_tasks_count: Optional[int] = None,
    max_retries_per_task: Optional[int] = None,
    max_worker_count: Optional[int] = None,
    target_task_run_status: Optional[str] = None,
    require_paths_exist: bool = False,
    submitter_name: Optional[str] = None,
    known_asset_paths: Collection[str] = [],
    from_gui: bool = False,
    print_function_callback: Callable[[str], None] = print,
    interactive_confirmation_callback: Optional[Callable[[str, bool], bool]] = None,
    hashing_progress_callback: Optional[Callable[[ProgressReportMetadata], bool]] = None,
    upload_progress_callback: Optional[Callable[[ProgressReportMetadata], bool]] = None,
    create_job_result_callback: Optional[Callable[[], bool]] = None,
) -> str:
    """
    Creates a job in the farm/queue configured as default for the
    workstation from the job bundle in the provided directory.

    A job bundle has the following directory structure:

    /template.json|yaml (required): An Open Job Description job template that specifies the work to be done. Job parameters
            are embedded here.
    /parameter_values.json|yaml (optional): If provided, these are parameter values for the job template and for
            the render farm. AWS Deadline Cloud-specific parameters are like "deadline:priority".
            Looks like:
            {
                "parameterValues": [
                    {"name": "<name>", "value": "<value>"},
                    ...
                ]
            }
    /asset_references.json|yaml (optional): If provided, these are references to the input and output assets
            of the job. Looks like:
            {
                "assetReferences": {
                    "inputs": {
                        "filenames": [
                            "/mnt/path/to/file.txt",
                            ...
                        ],
                        "directories": [
                            "/mnt/path/to/directory",
                            ...
                        ],
                    },
                    "outputs": {
                        "directories": [
                            "/mnt/path/to/output_directory",
                            ...
                        ],
                    }
                }
            }

    Args:
        job_bundle_dir (str): The directory containing the job bundle.
        job_parameters (List[Dict[str, Any]], optional): A list of job parameters in the following format:
            [{"name": "<name>", "value": "<value>"}, ...]
        name (str, optional): The name of the job to submit, replacing the name defined in the job bundle.
        queue_parameter_definitions (list[JobParameter], optional) A list of queue_parameters to use
                instead of retrieving queue_parameters from the queue with get_queue_parameter_definitions.
        job_attachments_file_system (str, optional): define which file system to use;
                (valid values: "COPIED", "VIRTUAL") instead of using the value in the config file.
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.
        priority (int, optional): explicit value for the priority of the job.
        max_failed_tasks_count (int, optional): explicit value for the maximum allowed failed tasks.
        max_retries_per_task (int, optional): explicit value for the maximum retries per task.
        max_worker_count (int, optional): explicit value for the max worker count of the job.
        target_task_run_status (str, optional): explicit value for the target task run status of the job.
                Valid values are "READY" or "SUSPENDED".
        require_paths_exist (bool, optional): Whether to require that all input paths exist.
        submitter_name (str, optional): Name of the application submitting the bundle.
        known_asset_paths (list[str], optional): A list of paths that should not generate
                warnings when outside storage profile locations. Defaults to an empty list.
        print_function_callback (Callable str -> None, optional): Callback to print messages produced in this function.
                By default calls print(), Can be replaced by click.echo or a logging function of choice.
        interactive_confirmation_callback (Callable [str, bool] -> bool): Callback arguments are (confirmation_message, default_response).
                This function should present the provided prompt, using default_response as the default value to respond with if the user
                does not make an explicit choice, and return True if the user wants to continue, False to cancel.
        hashing_progress_callback / upload_progress_callback / create_job_result_callback (Callable -> bool):
                Callbacks periodically called while hashing / uploading / waiting for job creation. If returns false,
                the operation will be cancelled. If return true, the operation continues. Default behavior for each
                is to not cancel the operation. hashing_progress_callback and upload_progress_callback both receive
                ProgressReport as a parameter, which can be used for projecting remaining time, as in done in the CLI.
    """

    if not submitter_name:
        submitter_name = "Custom"

    session_context["submitter-name"] = submitter_name

    # Ensure the job bundle doesn't contain files that resolve outside of the bundle directory
    validate_directory_symlink_containment(job_bundle_dir)

    # Read in the job template
    file_contents, file_type = read_yaml_or_json(job_bundle_dir, "template", required=True)

    # If requested, substitute the job name in the template
    if name is not None:
        template_obj = parse_yaml_or_json_content(
            file_contents, file_type, job_bundle_dir, "template"
        )
        template_obj["name"] = name
        if file_type == "YAML":
            file_contents = deadline_yaml_dump(template_obj)
        else:
            file_contents = json.dumps(template_obj)

    deadline = api.get_boto3_client("deadline", config=config)
    queue_id = get_setting("defaults.queue_id", config=config)
    farm_id = get_setting("defaults.farm_id", config=config)

    if job_attachments_file_system is None:
        job_attachments_file_system = get_setting(
            "defaults.job_attachments_file_system", config=config
        )

    queue = deadline.get_queue(
        farmId=farm_id,
        queueId=queue_id,
    )
    print_function_callback(f"Submitting to Queue: {queue['displayName']}\n")

    create_job_args: Dict[str, Any] = {
        "farmId": farm_id,
        "queueId": queue_id,
        "template": file_contents,
        "templateType": file_type,
        "priority": 50,
    }

    storage_profile_id = get_setting("settings.storage_profile_id", config=config)
    storage_profile = None
    if storage_profile_id:
        create_job_args["storageProfileId"] = storage_profile_id
        storage_profile = api.get_storage_profile_for_queue(
            farm_id, queue_id, storage_profile_id, deadline
        )

    # The job parameters
    job_bundle_parameters = read_job_bundle_parameters(job_bundle_dir)

    asset_references_obj = read_yaml_or_json_object(
        job_bundle_dir, "asset_references", required=False
    )
    asset_references = AssetReferences.from_dict(asset_references_obj)

    if queue_parameter_definitions is None:
        queue_parameter_definitions = api.get_queue_parameter_definitions(
            farmId=farm_id, queueId=queue_id
        )

    parameters = merge_queue_job_parameters(
        queue_id=queue_id,
        job_parameters=job_bundle_parameters,
        queue_parameters=queue_parameter_definitions,
    )

    apply_job_parameters(
        job_parameters,
        job_bundle_dir,
        parameters,
        asset_references,
    )
    app_parameters_formatted, job_parameters_formatted = split_parameter_args(
        parameters, job_bundle_dir
    )

    # Extend known_asset_paths with all paths that are treated as known. These are
    # paths provided explicitly by the call to submit the job bundle:
    #   * Paths in the known_asset_paths parameter to this function call
    #   * Paths contained inside the job bundle.
    #   * Paths configured in the locally configured storage profile as LOCAL (not SHARED).
    #   * Paths configured in the local config file settings.known_asset_paths
    #   * Paths provided within the job_parameters parameter to this function call
    # Paths that are treated as unknown (unless in one of the above categories). These can be
    # absolute paths referencing anywhere in the file system, not explicitly provided by the call,
    # so require that they be marked as known in the local configuration file or the associated
    # Storage Profile in the AWS account:
    #   * Paths provided in the job bundle via the parameter_values.json/.yaml file
    #   * Paths provided in the job bundle via the asset_references.json/.yaml files
    known_asset_paths = list(known_asset_paths) + [os.path.abspath(job_bundle_dir)]
    # Add the configured storage profile paths
    if storage_profile:
        known_asset_paths.extend(
            [
                fsl.path
                for fsl in storage_profile.fileSystemLocations
                if fsl.type == FileSystemLocationType.LOCAL
            ]
        )
    # Add the configured known asset paths
    configured_known_asset_paths = config_file.get_setting(
        "settings.known_asset_paths", config=config
    ).strip()
    if configured_known_asset_paths:
        known_asset_paths.extend(configured_known_asset_paths.split(os.pathsep))
    # Use the parameter names from job_parameters, but the values from parameters. If a value was provided
    # in job_parameters, it has been applied into parameters and normalized as necessary.
    known_parameter_names = {job_param.get("name") for job_param in job_parameters}
    for job_param in parameters:
        if job_param.get("type") == "PATH" and job_param.get("name") in known_parameter_names:
            job_param_value = job_param.get("value")
            if job_param_value:
                if job_param.get("objectType") == "FILE":
                    # If the job parameter is a file, use its directory as the known path. When collecting
                    # outputs for upload, only that directory is used, not the file path.
                    known_asset_paths.append(os.path.dirname(job_param_value))
                else:
                    known_asset_paths.append(job_param_value)

    # Filter known_asset_paths to remove any paths that have another one as a prefix. This can
    # reduce the amount of processing needed later, and produces a shorter warning message when presenting
    # to users.
    known_asset_paths = _filter_redundant_known_paths(known_asset_paths)

    # Hash and upload job attachments if there are any
    files_processed = False
    if asset_references and "jobAttachmentSettings" in queue:
        # Extend input_filenames with all the files in the input_directories
        missing_directories: set[str] = set()
        for directory in asset_references.input_directories:
            if not os.path.isdir(directory):
                if require_paths_exist:
                    missing_directories.add(directory)
                else:
                    logger.warning(
                        f"Input path '{directory}' does not exist. Adding to referenced paths."
                    )
                    asset_references.referenced_paths.add(directory)
                continue

            is_dir_empty = True
            for root, _, files in os.walk(directory):
                if not files:
                    continue
                is_dir_empty = False
                asset_references.input_filenames.update(
                    os.path.normpath(os.path.join(root, file)) for file in files
                )
            # Empty directories just become references since the current asset manifest spec
            # version cannot represent them.
            if is_dir_empty:
                logger.info(f"Input directory '{directory}' is empty. Adding to referenced paths.")
                asset_references.referenced_paths.add(directory)
        asset_references.input_directories.clear()

        if missing_directories:
            all_missing_directories = "\n\t".join(sorted(list(missing_directories)))
            misconfigured_directories_msg = (
                "Job submission contains misconfigured input directories and cannot be submitted."
                " All input directories must exist."
                f"\nNon-existent directories:\n\t{all_missing_directories}"
            )

            raise MisconfiguredInputsError(misconfigured_directories_msg)

        queue_role_session = api.get_queue_user_boto3_session(
            deadline=deadline,
            config=config,
            farm_id=farm_id,
            queue_id=queue_id,
            queue_display_name=queue["displayName"],
        )

        asset_manager = S3AssetManager(
            farm_id=farm_id,
            queue_id=queue_id,
            job_attachment_settings=JobAttachmentS3Settings(**queue["jobAttachmentSettings"]),
            session=queue_role_session,
        )

        upload_group = asset_manager.prepare_paths_for_upload(
            input_paths=sorted(asset_references.input_filenames),
            output_paths=sorted(asset_references.output_directories),
            referenced_paths=sorted(asset_references.referenced_paths),
            storage_profile=storage_profile,
            require_paths_exist=require_paths_exist,
        )

        if upload_group.asset_groups:
            # Generate warning message if needed
            asset_path_message, default_prompt_response = _generate_message_for_asset_paths(
                upload_group, storage_profile, known_asset_paths
            )

            if interactive_confirmation_callback is None:
                # In this case, no user prompt can be presented. The result of the function must
                # be the default that would be presented to the interactive prompt.
                print_function_callback(asset_path_message)
                if not default_prompt_response:
                    print_function_callback("\nJob submission canceled (user input not enabled).")
                    raise DeadlineOperationCanceled()
            elif config_file.str2bool(get_setting("settings.auto_accept", config=config)):
                if not default_prompt_response:
                    if from_gui:
                        # In the from_gui case, we present a prompt even though settings.auto_accept is enabled.
                        if not interactive_confirmation_callback(
                            asset_path_message + "Do you wish to proceed?", default_prompt_response
                        ):
                            print_function_callback("Job submission canceled (user input).")
                            raise UserInitiatedCancel()
                    else:
                        # In this case, no user prompt should be presented. The result of the function must
                        # be the default that would be presented to the interactive prompt.
                        print_function_callback(
                            f"{asset_path_message}\nJob submission canceled (settings.auto_accept enabled and there were unknown paths)."
                        )
                        raise DeadlineOperationCanceled()
                else:
                    print_function_callback(asset_path_message)
            else:
                if not interactive_confirmation_callback(
                    asset_path_message + "\nDo you wish to proceed?", default_prompt_response
                ):
                    print_function_callback("Job submission canceled (user input).")
                    raise UserInitiatedCancel()

            _, asset_manifests = _hash_attachments(
                asset_manager=asset_manager,
                asset_groups=upload_group.asset_groups,
                total_input_files=upload_group.total_input_files,
                total_input_bytes=upload_group.total_input_bytes,
                print_function_callback=print_function_callback,
                hashing_progress_callback=hashing_progress_callback,
            )

            attachment_settings = _upload_attachments(  # type: ignore
                asset_manager,
                asset_manifests,
                print_function_callback,
                upload_progress_callback,
                from_gui=from_gui,
            )
            attachment_settings["fileSystem"] = JobAttachmentsFileSystem(
                job_attachments_file_system
            )
            create_job_args["attachments"] = attachment_settings

            files_processed = True

    if not files_processed:
        # Call each callback once indicating nothing to do.
        if hashing_progress_callback is not None:
            hashing_progress_callback(
                ProgressReportMetadata(
                    status=ProgressStatus.PREPARING_IN_PROGRESS,
                    progress=0,
                    transferRate=0,
                    progressMessage="No files to hash",
                )
            )
        if upload_progress_callback is not None:
            upload_progress_callback(
                ProgressReportMetadata(
                    status=ProgressStatus.UPLOAD_IN_PROGRESS,
                    progress=0,
                    transferRate=0,
                    progressMessage="No files to upload",
                )
            )

    create_job_args.update(app_parameters_formatted)

    if job_parameters_formatted:
        create_job_args["parameters"] = job_parameters_formatted

    if priority is not None:
        create_job_args["priority"] = priority
    if max_worker_count is not None:
        create_job_args["maxWorkerCount"] = max_worker_count
    if max_failed_tasks_count is not None:
        create_job_args["maxFailedTasksCount"] = max_failed_tasks_count
    if max_retries_per_task is not None:
        create_job_args["maxRetriesPerTask"] = max_retries_per_task
    if target_task_run_status is not None:
        create_job_args["targetTaskRunStatus"] = target_task_run_status

    if logging.DEBUG >= logger.getEffectiveLevel():
        logger.debug(json.dumps(create_job_args, indent=1))

    api.get_deadline_cloud_library_telemetry_client().record_event(
        event_type="com.amazon.rum.deadline.submission",
        event_details={"submitter_name": submitter_name},
        from_gui=from_gui,
    )

    create_job_response = deadline.create_job(**create_job_args)
    logger.debug("CreateJob Response %r", create_job_response)

    if create_job_response and "jobId" in create_job_response:
        job_id = create_job_response["jobId"]
        print_function_callback("Waiting for Job to be created...")

        # If using the default config, set the default job id so it holds the
        # most-recently submitted job.
        if config is None:
            set_setting("defaults.job_id", job_id)

        def _default_create_job_result_callback() -> bool:
            return True

        if not create_job_result_callback:
            create_job_result_callback = _default_create_job_result_callback

        success, status_message = wait_for_create_job_to_complete(
            farm_id,
            queue_id,
            job_id,
            deadline,
            create_job_result_callback,
        )

        api.get_deadline_cloud_library_telemetry_client().record_event(
            event_type="com.amazon.rum.deadline.create_job",
            event_details={"is_success": success},
            from_gui=from_gui,
        )

        if not success:
            raise DeadlineOperationError(status_message)

        print_function_callback("Submitted job bundle:")
        print_function_callback(f"   {job_bundle_dir}")
        print_function_callback(status_message + f"\n{job_id}")

        return job_id
    else:
        raise DeadlineOperationError("CreateJob response was empty, or did not contain a Job ID.")


def wait_for_create_job_to_complete(
    farm_id: str,
    queue_id: str,
    job_id: str,
    deadline_client: BaseClient,
    continue_callback: Callable,
) -> Tuple[bool, str]:
    """
    Wait until a job exits the CREATE_IN_PROGRESS state.
    """

    delay_sec = 5  # Time to wait between checks in seconds.
    max_attempts = 60
    creating_statuses = {
        "CREATE_IN_PROGRESS",
    }
    failure_statuses = {"CREATE_FAILED"}

    for attempt in range(max_attempts):
        logger.debug(
            f"Waiting for creation of {job_id} to complete...attempt {attempt} of {max_attempts}"
        )

        if not continue_callback():
            raise CreateJobWaiterCanceled()

        job = deadline_client.get_job(jobId=job_id, queueId=queue_id, farmId=farm_id)

        current_status = job["lifecycleStatus"] if "lifecycleStatus" in job else job["state"]
        if current_status in creating_statuses:
            time.sleep(delay_sec)
        elif current_status in failure_statuses:
            return False, job["lifecycleStatusMessage"]
        else:
            return True, job["lifecycleStatusMessage"]

    raise TimeoutError(
        f"Timed out after {delay_sec * max_attempts} seconds while waiting for Job to be created: {job_id}"
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = [
    "adaptors",
    "create_job_history_bundle_dir",
    "read_job_bundle_parameters",
    "apply_job_parameters",
    "deadline_yaml_dump",
]

import datetime
import glob
import os

from ..config import get_setting
from ._yaml import deadline_yaml_dump
from .parameters import apply_job_parameters, read_job_bundle_parameters


def create_job_history_bundle_dir(submitter_name: str, job_name: str) -> str:
    """
    Creates a new directory in the configured directory
    settings.job_history_dir, in which to place a new
    job bundle for submission.

    The directory will look like
      <job_history_dir>/YYYY-mm/YYYY-mm-ddTHH-##-<submitter_name>-<job_name>
    """
    job_history_dir = str(get_setting("settings.job_history_dir"))
    job_history_dir = os.path.expanduser(job_history_dir)

    # Clean the submitter_name's characters
    submitter_name_cleaned = "".join(
        char for char in submitter_name if char.isalnum() or char in " -_"
    )

    # Clean the job_name's characters and truncate for the filename
    job_name_cleaned = "".join(char for char in job_name if char.isalnum() or char in " -_")
    job_name_cleaned = job_name_cleaned[:128]

    timestamp = datetime.datetime.now()
    month_tag = timestamp.strftime("%Y-%m")
    date_tag = timestamp.strftime("%Y-%m-%d")

    month_dir = os.path.join(job_history_dir, month_tag)
    if not os.path.isdir(month_dir):
        os.makedirs(month_dir)

    # Index the files so they sort in order of submission
    number = 1
    existing_dirs = sorted(glob.glob(os.path.join(month_dir, f"{date_tag}-*")))
    if existing_dirs:
        latest_dir = existing_dirs[-1]
        number = int(os.path.basename(latest_dir)[len(date_tag) + 1 :].split("-", 1)[0]) + 1

    result = os.path.join(
        month_dir, f"{date_tag}-{number:02}-{submitter_name_cleaned}-{job_name_cleaned}"
    )
    os.makedirs(result)
    return result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
import concurrent.futures

import logging
import os
from pathlib import Path, PurePosixPath
from typing import Dict, List, Tuple
from math import trunc
from deadline.client.cli._groups.click_logger import ClickLogger
from deadline.client.config import config_file
from deadline.client.exceptions import NonValidInputError
from deadline.job_attachments.asset_manifests.base_manifest import (
    BaseAssetManifest,
    BaseManifestPath,
)
from deadline.job_attachments.caches.hash_cache import HashCache
from deadline.job_attachments.models import AssetRootManifest, FileStatus, ManifestDiff
from deadline.job_attachments.upload import S3AssetManager


def diff_manifest(
    asset_manager: S3AssetManager,
    asset_root_manifest: AssetRootManifest,
    manifest: str,
    update: bool,
) -> List[(Tuple[FileStatus, BaseManifestPath])]:
    """
    Gets the file paths in specified manifest if the contents of file have changed since its last snapshot.
    Returns a list of FileStatus and BaseManifestPath
    """
    manifest_dir_name: str = os.path.basename(manifest)
    root_path: str = asset_root_manifest.root_path
    input_paths: List[Path] = []

    asset_manifest = asset_root_manifest.asset_manifest
    if asset_manifest is None:
        raise NonValidInputError("Manifest object not found, please check input manifest. ")

    for base_manifest_path in asset_manifest.paths:
        if base_manifest_path.path.startswith(manifest_dir_name):
            # skip the manifest folder, or else every upload will need an update after a previous change
            continue
        input_paths.append(Path(root_path, base_manifest_path.path))

    return find_file_with_status(
        asset_manager=asset_manager,
        input_paths=input_paths,
        root_path=root_path,
        update=update,
        statuses=[FileStatus.NEW, FileStatus.MODIFIED],
    )


def find_file_with_status(
    asset_manager: S3AssetManager,
    input_paths: List[Path],
    root_path: str,
    update: bool,
    statuses: List[FileStatus],
) -> List[(Tuple[FileStatus, BaseManifestPath])]:
    """
    Checks a manifest file, compares it to specified root directory or manifest of files with the local hash cache, and finds files that match the specified statuses.
    Returns a list of tuples containing the file information, and its corresponding file status.
    """
    cache_config: str = config_file.get_cache_directory()

    with HashCache(cache_config) as hash_cache:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(
                    asset_manager._process_input_path,
                    path=path,
                    root_path=root_path,
                    hash_cache=hash_cache,
                    update=update,
                ): path
                for path in input_paths
            }
            status_paths: List[tuple] = []
            for future in concurrent.futures.as_completed(futures):
                (file_status, _, manifestPath) = future.result()
                if file_status in statuses:
                    status_paths.append((file_status, manifestPath))

            return status_paths


def compare_manifest(
    reference_manifest: BaseAssetManifest, compare_manifest: BaseAssetManifest
) -> List[(Tuple[FileStatus, BaseManifestPath])]:
    """
    Compares two manifests, reference_manifest acting as the base, and compare_manifest acting as manifest with changes.
    Returns a list of FileStatus and BaseManifestPath
    """
    reference_dict: Dict[str, BaseManifestPath] = {
        manifest_path.path: manifest_path for manifest_path in reference_manifest.paths
    }
    compare_dict: Dict[str, BaseManifestPath] = {
        manifest_path.path: manifest_path for manifest_path in compare_manifest.paths
    }

    differences: List[(Tuple[FileStatus, BaseManifestPath])] = []

    # Find new files
    for file_path, manifest_path in compare_dict.items():
        if file_path not in reference_dict:
            differences.append((FileStatus.NEW, manifest_path))
        elif reference_dict[file_path].hash != manifest_path.hash:
            differences.append((FileStatus.MODIFIED, manifest_path))
        else:
            differences.append((FileStatus.UNCHANGED, manifest_path))

    # Find deleted files
    for file_path, manifest_path in reference_dict.items():
        if file_path not in compare_dict:
            differences.append((FileStatus.DELETED, manifest_path))

    return differences


def _fast_file_list_to_manifest_diff(
    root: str,
    current_files: List[str],
    diff_manifest: BaseAssetManifest,
    logger: ClickLogger,
    return_root_relative_path: bool = True,
) -> List[Tuple[str, FileStatus]]:
    """
    Perform a fast difference of the current list of files to a previous manifest to diff against using time stamps and file sizes.
    :param root: Root folder of files to diff against.
    :param current_files: List of files to compare with.
    :param diff_manifest: Manifest containing files to diff against.
    :param return_root_relative_path: File Path to return, either relative to root or full.
    :param logger: logger.
    :return List[Tuple[str, FileStatus]]: List of Tuple containing the file path and FileStatus pair.
    """

    # Select either relative or absolut path for results.
    def select_path(full_path: str, relative_path: str, return_root_relative_path: bool):
        return relative_path if return_root_relative_path else full_path

    changed_paths: List[Tuple[str, FileStatus]] = []
    input_files_map: Dict[str, BaseManifestPath] = {}
    for input_file in diff_manifest.paths:
        # Normalize paths so we can compare different OSes
        normalized_path = Path(os.path.normpath(input_file.path)).as_posix()
        input_files_map[normalized_path] = input_file

    # Iterate for each file that we found in glob.
    root_relative_paths: List[str] = []
    for local_file in current_files:
        # Get the file's time stamp and size. We want to compare both.
        # From enabling CRT, sometimes timestamp update can fail.
        local_file_path = Path(local_file)
        file_stat = local_file_path.stat()

        # Compare the glob against the relative path we store in the manifest.
        # Save it to a list so we can look for deleted files.
        root_relative_path = str(PurePosixPath(*local_file_path.relative_to(root).parts).as_posix())
        root_relative_paths.append(root_relative_path)

        return_path = select_path(
            full_path=local_file,
            relative_path=root_relative_path,
            return_root_relative_path=return_root_relative_path,
        )
        if root_relative_path not in input_files_map:
            # This is a new file
            logger.echo(f"Found difference at: {root_relative_path}, Status: FileStatus.NEW")
            changed_paths.append((return_path, FileStatus.NEW))
        else:
            # This is a modified file, compare with manifest relative timestamp.
            input_file = input_files_map[root_relative_path]
            # Check file size first as it is easier to test. Usually modified files will also have size diff.
            if file_stat.st_size != input_file.size:
                changed_paths.append((return_path, FileStatus.MODIFIED))
                logger.echo(
                    f"Found size difference at: {root_relative_path}, Status: FileStatus.MODIFIED"
                )
            # Check file mtime, allow 1 microsecond diff to prevent false positive
            # utime set from microsecond to nanosecond conversion could create 1 microsecond diff upon division
            elif abs(trunc(file_stat.st_mtime_ns / 1000) - input_file.mtime) > 1:
                changed_paths.append((return_path, FileStatus.MODIFIED))
                logger.echo(
                    f"Found time difference at: {root_relative_path}, Status: FileStatus.MODIFIED"
                )

    # Find deleted files. Manifest store files in relative form.
    for manifest_file_path in diff_manifest.paths:
        if manifest_file_path.path not in root_relative_paths:
            full_path = os.path.join(root, manifest_file_path.path)
            return_path = select_path(
                full_path=full_path,
                relative_path=manifest_file_path.path,
                return_root_relative_path=return_root_relative_path,
            )
            changed_paths.append((return_path, FileStatus.DELETED))
    return changed_paths


def pretty_print_cli(root: str, all_files: List[str], manifest_diff: ManifestDiff):
    """
    Prints to command line a formatted file tree structure with corresponding file statuses
    """

    # ASCII characters for the tree structure
    PIPE = "│"
    HORIZONTAL = "──"
    ELBOW = "└"
    TEE = "├"
    SPACE = "    "

    # ANSI escape sequences for colors
    COLORS = {
        "MODIFIED": "\033[93m",  # yellow
        "NEW": "\033[92m",  # green
        "DELETED": "\033[91m",  # red
        "UNCHANGED": "\033[90m",  # grey
        "RESET": "\033[0m",  # base color
        "DIRECTORY": "\033[80m",  # grey
    }

    # Tooltips:
    TOOLTIPS = {
        FileStatus.NEW: " +",  # added files
        FileStatus.DELETED: " -",  # deleted files
        FileStatus.MODIFIED: " M",  # modified files
        FileStatus.UNCHANGED: "",  # unchanged files
    }

    class ColorFormatter(logging.Formatter):
        def format(self, record):
            message = super().format(record)
            return f"{message}"

    # Configure logger
    formatter = ColorFormatter("")
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger = logging.getLogger(__name__)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    logger.propagate = False

    def print_tree(directory_tree, prefix=""):
        sorted_entries = sorted(directory_tree.items())

        for i, (entry, subtree) in enumerate(sorted_entries, start=1):
            is_last_entry = i == len(sorted_entries)
            symbol = ELBOW + HORIZONTAL if is_last_entry else TEE + HORIZONTAL
            is_dir = isinstance(subtree, dict)
            color = COLORS["DIRECTORY"] if is_dir else COLORS[subtree.name]
            tooltip = TOOLTIPS[FileStatus.UNCHANGED] if is_dir else TOOLTIPS[subtree]

            message = f"{prefix}{symbol}{color}{entry}{tooltip}{COLORS['RESET']}{os.path.sep if is_dir else ''}"
            logger.info(message)

            if is_dir:
                new_prefix = prefix + (SPACE if is_last_entry else PIPE + SPACE)
                print_tree(subtree, new_prefix)

        if not directory_tree:
            symbol = ELBOW + HORIZONTAL
            message = f"{prefix}{symbol}{COLORS['UNCHANGED']}. {COLORS['RESET']}"
            logger.info(message)

    def get_file_status(file: str, manifest_diff: ManifestDiff):
        print(file)
        if file in manifest_diff.new:
            return FileStatus.NEW
        elif file in manifest_diff.modified:
            return FileStatus.MODIFIED
        elif file in manifest_diff.deleted:
            return FileStatus.DELETED
        else:
            # Default, not in any diff list.
            return FileStatus.UNCHANGED

    def build_directory_tree(all_files: List[str]) -> Dict[str, dict]:
        directory_tree: dict = {}

        def add_to_tree(path, status):
            parts = str(path).split(os.path.sep)
            current_level = directory_tree
            for i, part in enumerate(parts):
                if i == len(parts) - 1:
                    current_level[part] = status
                else:
                    current_level = current_level.setdefault(part, {})

        for file in all_files:
            print(f"{file} {root}")
            relative_path = str(Path(file).relative_to(root))
            add_to_tree(
                relative_path,
                get_file_status(relative_path, manifest_diff),
            )
        return directory_tree

    directory_tree = build_directory_tree(all_files)
    print_tree(directory_tree)
    logger.info("")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline bundle` commands.
"""

from __future__ import annotations

import json
import logging
import sys
import re
from typing import Any, Optional

import click
from botocore.exceptions import ClientError

from ... import api
from ...config import config_file
from ....job_attachments.exceptions import (
    AssetSyncError,
    AssetSyncCancelledError,
    MisconfiguredInputsError,
)
from ....job_attachments.models import JobAttachmentsFileSystem

from ...exceptions import DeadlineOperationError, CreateJobWaiterCanceled
from .._common import (
    _apply_cli_options_to_config,
    _handle_error,
    _ProgressBarCallbackManager,
)
from .._main import main
from ._sigint_handler import SigIntHandler

logger = logging.getLogger(__name__)

# Set up the signal handler for handling Ctrl + C interruptions.
sigint_handler = SigIntHandler()


@main.group(name="bundle")
@_handle_error
def cli_bundle():
    """
    Commands to work with Open Job Description job bundles.
    """


# Latin alphanumeric, starting with a letter
_openjd_identifier_regex = r"(?-m:^[A-Za-z_][A-Za-z0-9_]*\Z)"


def validate_parameters(ctx, param, value):
    """
    Validate provided --parameter values, ensuring that they are in the format "ParamName=Value", and convert them to a dict with the
    following format:
        [{"name": "<name>", "value": "<value>"}, ...]
    """
    parameters_split = []
    for parameter in value:
        regex_match = re.match("([^=]+)=(.*)", parameter)
        if not regex_match:
            raise click.BadParameter(
                f'Parameters must be provided in the format "ParamName=Value". Invalid parameter: {parameter}'
            )

        if not re.match(_openjd_identifier_regex, regex_match[1]):
            raise click.BadParameter(
                f"Parameter names must be alphanumeric Open Job Description identifiers. Invalid parameter name: {regex_match[1]}"
            )

        parameters_split.append({"name": regex_match[1], "value": regex_match[2]})

    return parameters_split


def _interactive_confirmation_prompt(message: str, default_response: bool) -> bool:
    """
    Callback to decide if submission should continue or be canceled. Returns True to continue, False to cancel.

    Args:
        warning_message (str): The warning message to display.
        default_response (bool): The default to present as the response (True to continue, False to cancel).
    """
    return click.confirm(
        message,
        default=default_response,
    )


@cli_bundle.command(name="submit")
@click.option(
    "-p",
    "--parameter",
    multiple=True,
    callback=validate_parameters,
    help='The values for the job template\'s parameters. Can be provided as key-value pairs, inline JSON strings, or as paths to a JSON or YAML document. If provided more than once, the values are combined in the order that they appear. Examples: --parameter MyParam=5 -p file://parameter_file.json -p \'{"MyParam": "5"}\'',
)
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--storage-profile-id", help="The storage profile to use.")
@click.option("--name", help="The job name to use in place of the one in the job bundle.")
@click.option(
    "--priority",
    type=int,
    default=50,
    help="The priority of the job. Jobs with a higher priority run first.",
)
@click.option(
    "--max-failed-tasks-count",
    type=int,
    help="The maximum number of failed tasks before the job is marked as failed.",
)
@click.option(
    "--max-retries-per-task",
    type=int,
    help="The maximum number of times to retry a task before it is marked as failed.",
)
@click.option(
    "--max-worker-count",
    type=int,
    help="The max worker count of the job.",
)
@click.option(
    "--target-task-run-status",
    type=click.Choice(["READY", "SUSPENDED"], case_sensitive=False),
    help="The target task run status for the job. READY means tasks will start immediately, "
    "SUSPENDED means tasks will be created but not start until manually resumed.",
)
@click.option(
    "--job-attachments-file-system",
    help="The method workers use to access job attachments. "
    "COPIED means to copy files to the worker and VIRTUAL means to load "
    "files as needed from a virtual file system. If VIRTUAL is selected "
    "but not supported by a worker, it will fallback to COPIED.",
    type=click.Choice([e.value for e in JobAttachmentsFileSystem]),
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@click.option(
    "--require-paths-exist",
    is_flag=True,
    help="Return an error if any input files are missing.",
)
@click.option(
    "--submitter-name",
    type=click.STRING,
    help="Name of the application submitting the bundle.",
)
@click.option(
    "--known-asset-path",
    multiple=True,
    help="Path that should not generate warnings when outside storage profile locations. "
    "Can be specified multiple times for different paths.",
)
@click.argument("job_bundle_dir")
@_handle_error
def bundle_submit(
    job_bundle_dir,
    job_attachments_file_system,
    parameter,
    known_asset_path,
    name,
    priority,
    max_failed_tasks_count,
    max_retries_per_task,
    max_worker_count,
    target_task_run_status,
    require_paths_exist,
    submitter_name,
    **args,
):
    """
    Submits an Open Job Description job bundle.
    """
    # Apply the CLI args to the config
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    hash_callback_manager = _ProgressBarCallbackManager(length=100, label="Hashing Attachments")
    upload_callback_manager = _ProgressBarCallbackManager(length=100, label="Uploading Attachments")

    def _check_create_job_wait_canceled() -> bool:
        return sigint_handler.continue_operation

    try:
        job_id = api.create_job_from_job_bundle(
            job_bundle_dir=job_bundle_dir,
            job_parameters=parameter,
            name=name,
            job_attachments_file_system=job_attachments_file_system,
            config=config,
            priority=priority,
            max_failed_tasks_count=max_failed_tasks_count,
            max_retries_per_task=max_retries_per_task,
            max_worker_count=max_worker_count,
            target_task_run_status=target_task_run_status,
            hashing_progress_callback=hash_callback_manager.callback,
            upload_progress_callback=upload_callback_manager.callback,
            create_job_result_callback=_check_create_job_wait_canceled,
            print_function_callback=click.echo,
            interactive_confirmation_callback=_interactive_confirmation_prompt,
            require_paths_exist=require_paths_exist,
            submitter_name=submitter_name or "CLI",
            known_asset_paths=known_asset_path,
        )

        # Check Whether the CLI options are modifying any of the default settings that affect
        # the job id. If not, we'll save the job id submitted as the default job id.
        # If the submission is canceled by the user job_id will be None, so ignore this case as well.
        if (
            args.get("profile") is None
            and args.get("farm_id") is None
            and args.get("queue_id") is None
            and args.get("storage_profile_id") is None
        ):
            config_file.set_setting("defaults.job_id", job_id)

    except AssetSyncCancelledError as exc:
        if sigint_handler.continue_operation:
            raise DeadlineOperationError(f"Job submission unexpectedly canceled:\n{exc}") from exc
        else:
            click.echo("Job submission canceled.")
            sys.exit(1)
    except AssetSyncError as exc:
        raise DeadlineOperationError(f"Failed to upload job attachments:\n{exc}") from exc
    except CreateJobWaiterCanceled as exc:
        if sigint_handler.continue_operation:
            raise DeadlineOperationError(
                f"Unexpectedly canceled during wait for final status of CreateJob:\n{exc}"
            ) from exc
        else:
            click.echo("Canceled waiting for final status of CreateJob.")
            sys.exit(1)
    except ClientError as exc:
        raise DeadlineOperationError(
            f"Failed to submit the job bundle to AWS Deadline Cloud:\n{exc}"
        ) from exc
    except MisconfiguredInputsError as exc:
        click.echo(str(exc))
        click.echo("Job submission canceled.")
        sys.exit(1)
    except Exception as exc:
        api.get_deadline_cloud_library_telemetry_client().record_error(
            event_details={"exception_scope": "on_submit"},
            exception_type=str(type(exc)),
        )
        raise


@cli_bundle.command(name="gui-submit")
@click.argument("job_bundle_dir", required=False)
@click.option(
    "--browse",
    is_flag=True,
    help="Opens a folder browser to select a bundle.",
)
@click.option(
    "--install-gui",
    is_flag=True,
    help="Installs GUI dependencies if they are not installed already",
)
@click.option(
    "--submitter-name",
    type=click.STRING,
    help="Name of the application submitting the bundle. If a name is specified, the GUI will automatically close after submitting the job.",
)
@click.option(
    "--output",
    type=click.Choice(
        ["verbose", "json"],
        case_sensitive=False,
    ),
    default="verbose",
    help="Specifies the output format of the messages printed to stdout.\n"
    "VERBOSE: Displays messages in a human-readable text format.\n"
    "JSON: Displays messages in JSON line format, so that the info can be easily "
    "parsed/consumed by custom scripts.",
)
@click.option(
    "--known-asset-path",
    multiple=True,
    help="Path that should not generate warnings when outside storage profile locations. "
    "Can be specified multiple times for different paths.",
)
@_handle_error
def bundle_gui_submit(
    job_bundle_dir, browse, output, install_gui, submitter_name, known_asset_path, **args
):
    """
    Opens a GUI to submit an Open Job Description job bundle.
    """
    from ...ui import gui_context_for_cli

    with gui_context_for_cli(automatically_install_dependencies=install_gui) as app:
        from ...ui.job_bundle_submitter import show_job_bundle_submitter

        if not job_bundle_dir and not browse:
            raise DeadlineOperationError(
                "Specify a job bundle directory or run the bundle command with the --browse flag"
            )
        output = output.lower()

        submitter = show_job_bundle_submitter(
            input_job_bundle_dir=job_bundle_dir,
            browse=browse,
            submitter_name=submitter_name,
            known_asset_paths=known_asset_path,
        )

        if not submitter:
            return

        submitter.show()

        app.exec()

        _print_response(
            output=output,
            job_bundle_dir=job_bundle_dir,
            job_history_bundle_dir=submitter.job_history_bundle_dir,
            job_id=submitter.job_id,
        )


def _print_response(
    output: str,
    job_bundle_dir: str,
    job_history_bundle_dir: Optional[str],
    job_id: Optional[str],
):
    if output == "json":
        if job_id:
            response: dict[str, Any] = {
                "status": "SUBMITTED",
                "jobId": job_id,
                "jobHistoryBundleDirectory": job_history_bundle_dir,
            }
            click.echo(json.dumps(response))
        else:
            click.echo(json.dumps({"status": "CANCELED"}))
    else:
        if job_id:
            click.echo("Submitted job bundle:")
            click.echo(f"   {job_bundle_dir}")
            click.echo(f"Job ID: {job_id}")
        else:
            click.echo("Job submission canceled.")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from .asset_manifest import AssetManifest, ManifestModel, ManifestPath

__all__ = ["ManifestModel", "ManifestPath", "AssetManifest"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
UI widgets for the scene settings tab.
"""

from __future__ import annotations

import os
from logging import getLogger
from typing import Any

from qtpy.QtCore import Signal  # type: ignore
from qtpy.QtWidgets import (  # type: ignore
    QVBoxLayout,
    QHBoxLayout,
    QWidget,
    QFileDialog,
    QPushButton,
    QSpacerItem,
    QSizePolicy,
    QMessageBox,
)

from ..dataclasses import JobBundleSettings
from .openjd_parameters_widget import OpenJDParametersWidget
from ...job_bundle.submission import AssetReferences
from ...job_bundle.loader import read_yaml_or_json_object, validate_directory_symlink_containment
from ...job_bundle.parameters import read_job_bundle_parameters
from ...config import config_file

logger = getLogger(__name__)


class JobBundleSettingsWidget(QWidget):
    """
    Widget containing job setup specific to CLI jobs.

    Signals:
        parameter_changed: This is sent whenever a parameter value in the widget changes. The message
            is a copy of the parameter definition with the "value" key containing the new value.

    Args:
        initial_settings (CliJobSettings): dataclass containing the job-specific settings.
        parent: The parent Qt Widget.
    """

    parameter_changed = Signal(dict)

    def __init__(self, initial_settings: JobBundleSettings, parent=None):
        super().__init__(parent=parent)

        self.parent = parent

        self.param_layout = QVBoxLayout()

        self._build_ui(initial_settings)

    def _build_ui(self, initial_settings: JobBundleSettings):
        self.input_job_bundle_dir = initial_settings.input_job_bundle_dir

        layout = QVBoxLayout(self)

        if initial_settings.browse_enabled:
            btnBox = QHBoxLayout()
            self.load_bundle_button = QPushButton("Load a different job bundle")
            self.load_bundle_button.clicked.connect(self.on_load_bundle)
            btnBox.addWidget(self.load_bundle_button)
            btnBox.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
            layout.addLayout(btnBox)

        layout.addLayout(self.param_layout)
        self.refresh_ui(initial_settings)

    def refresh_ui(self, settings: JobBundleSettings):
        # Clear the layout
        for i in reversed(range(self.param_layout.count())):
            item = self.param_layout.takeAt(i)
            item.widget().deleteLater()

        self.parameters_widget = OpenJDParametersWidget(
            parameter_definitions=settings.parameters, parent=self
        )
        self.param_layout.addWidget(self.parameters_widget)
        self.parameters_widget.parameter_changed.connect(
            lambda message: self.parameter_changed.emit(message)
        )

    def on_load_bundle(self):
        """
        Browse and load the selected submission bundle
        """
        # Open the file picker dialog
        bundle_path = os.path.expanduser(config_file.get_setting("settings.job_history_dir"))
        input_job_bundle_dir = QFileDialog.getExistingDirectory(
            self, "Choose job bundle directory", bundle_path
        )
        if not input_job_bundle_dir:
            return

        # Update job bundle directory path
        self.input_job_bundle_dir = input_job_bundle_dir

        # Warn the user if the Job Bundle could not be loaded
        try:
            validate_directory_symlink_containment(input_job_bundle_dir)

            asset_references_obj = (
                read_yaml_or_json_object(input_job_bundle_dir, "asset_references", False) or {}
            )
            asset_references = AssetReferences.from_dict(asset_references_obj)

            # Load the template to get the bundle name
            template = read_yaml_or_json_object(input_job_bundle_dir, "template", True)
            name = (
                template.get("name", "Job bundle submission")
                if template
                else "Job bundle submission"
            )
            job_settings = JobBundleSettings(input_job_bundle_dir=input_job_bundle_dir, name=name)
            job_settings.parameters = read_job_bundle_parameters(input_job_bundle_dir)

        except Exception as e:
            msg = str(e)
            QMessageBox.warning(self, "Could not load job bundle", msg)  # type: ignore[call-arg]
            logger.warning(msg)
            return

        if hasattr(self.parent, "refresh"):
            self.parent.refresh(
                job_settings=job_settings,
                auto_detected_attachments=asset_references,
                attachments=None,
                load_new_bundle=True,
            )

    def update_settings(self, settings: JobBundleSettings):
        """
        Update a settings object with the latest values.
        """
        settings.input_job_bundle_dir = self.input_job_bundle_dir
        settings.parameters = self.parameters_widget.get_parameters()

    def get_parameters(self):
        """
        Returns a list of OpenJD parameter definition dicts with
        a "value" key filled from the widget.
        """
        return self.parameters_widget.get_parameters()

    def set_parameter_value(self, parameter: dict[str, Any]):
        """
        Given an OpenJD parameter definition with a "value" key,
        set the parameter value in the widget.

        If the parameter value cannot be set, raises a KeyError.
        """
        self.parameters_widget.set_parameter_value(parameter)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

__all__ = ["get_queue_parameter_definitions"]

import yaml

from .. import api
from ._list_apis import _call_paginated_deadline_list_api
from ._session import get_boto3_client
from ..exceptions import DeadlineOperationError
from ..job_bundle.parameters import (
    JobParameter,
    get_ui_control_for_parameter_definition,
    parameter_definition_difference,
    validate_job_parameter,
)


@api.record_function_latency_telemetry_event()
def get_queue_parameter_definitions(
    *, farmId: str, queueId: str, config=None
) -> list[JobParameter]:
    """
    This gets all the queue parameters definitions from the specified Queue. It does so
    by getting all the full templates for queue environments, and then combining
    them equivalently to the Deadline Cloud service logic.
    """
    deadline = get_boto3_client("deadline", config=config)
    response = _call_paginated_deadline_list_api(
        deadline.list_queue_environments,
        "environments",
        farmId=farmId,
        queueId=queueId,
    )
    queue_environments = sorted(
        (
            deadline.get_queue_environment(
                farmId=farmId,
                queueId=queueId,
                queueEnvironmentId=queue_env["queueEnvironmentId"],
            )
            for queue_env in response["environments"]
        ),
        key=lambda queue_env: queue_env["priority"],
    )
    queue_environment_templates = [
        yaml.safe_load(queue_env["template"]) for queue_env in queue_environments
    ]

    queue_parameters_definitions: dict[str, JobParameter] = {}
    for template in queue_environment_templates:
        for parameter in template.get("parameterDefinitions", []):
            parameter = validate_job_parameter(parameter, type_required=True, default_required=True)

            # If there is no group label, set it to the name of the Queue Environment
            if not parameter.get("userInterface", {}).get("groupLabel"):
                if "userInterface" not in parameter:
                    parameter["userInterface"] = {
                        "control": get_ui_control_for_parameter_definition(parameter)
                    }
                parameter["userInterface"]["groupLabel"] = (
                    f"Queue Environment: {template['environment']['name']}"
                )
            existing_parameter = queue_parameters_definitions.get(parameter["name"])
            if existing_parameter:
                differences = parameter_definition_difference(existing_parameter, parameter)
                if differences:
                    raise DeadlineOperationError(
                        f"Job template parameter {parameter['name']} is duplicated across queue environments with mismatched fields:\n"
                        + " ".join(differences)
                    )
            else:
                queue_parameters_definitions[parameter["name"]] = parameter

    return list(queue_parameters_definitions.values())



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides functionality for logging in or out of the AWS Profile that is
configured for AWS Deadline Cloud to use on the local workstation.
"""

from configparser import ConfigParser
from logging import getLogger
from typing import Callable, Optional
import subprocess
import sys

from ._session import (
    get_credentials_source,
    check_authentication_status,
    AwsCredentialsSource,
    AwsAuthenticationStatus,
)
from . import _session
from .. import api
from ..config import get_setting
from ..exceptions import DeadlineOperationError
import time

logger = getLogger(__name__)


class UnsupportedProfileTypeForLoginLogout(DeadlineOperationError):
    pass


def _login_deadline_cloud_monitor(
    on_pending_authorization: Optional[Callable],
    on_cancellation_check: Optional[Callable],
    config: Optional[ConfigParser] = None,
):
    # Deadline Cloud monitor writes the absolute path to itself to the config file
    deadline_cloud_monitor_path = get_setting("deadline-cloud-monitor.path", config=config)
    profile_name = get_setting("defaults.aws_profile_name", config=config)
    args = [deadline_cloud_monitor_path, "login", "--profile", profile_name]

    # Open Deadline Cloud monitor, non-blocking the user will keep Deadline Cloud monitor running in the background.
    try:
        if sys.platform.startswith("win"):
            # We don't hookup to stdin but do this to avoid issues on windows
            # See https://docs.python.org/3/library/subprocess.html#subprocess.STARTUPINFO.lpAttributeList
            p = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, stdin=subprocess.PIPE
            )
        else:
            p = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL
            )
    except FileNotFoundError:
        raise DeadlineOperationError(
            f"Could not find Deadline Cloud monitor at {deadline_cloud_monitor_path}. "
            f"Please ensure Deadline Cloud monitor is installed correctly and set up the {profile_name} profile again."
        )
    if on_pending_authorization:
        on_pending_authorization(
            credentials_source=AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN
        )
    # And wait for the user to complete login
    while True:
        # Deadline Cloud monitor is a GUI app that will keep on running
        # So we sit here and test that profile for validity until it works
        if check_authentication_status(config) == AwsAuthenticationStatus.AUTHENTICATED:
            return f"Deadline Cloud monitor profile: {profile_name}"
        if on_cancellation_check:
            # Check if the UI has signaled a cancel
            if on_cancellation_check():
                p.kill()
                raise Exception()
        if p.poll():
            # Deadline Cloud monitor has stopped, we assume it returned us an error on one line on stderr
            # but let's be specific about Deadline Cloud monitor failing incase the error is non-obvious
            # and let's tack on stdout incase it helps
            err_prefix = (
                f"Deadline Cloud monitor was not able to log into the {profile_name} profile:"
            )
            out = p.stdout.read().decode("utf-8") if p.stdout else ""
            raise DeadlineOperationError(f"{err_prefix}\n{out}")

        time.sleep(0.5)


@api.record_function_latency_telemetry_event()
def login(
    on_pending_authorization: Optional[Callable],
    on_cancellation_check: Optional[Callable],
    config: Optional[ConfigParser] = None,
) -> str:
    """
    For AWS profiles created by Deadline Cloud monitor, logs in to provide access to Deadline Cloud.

    Args:
        on_pending_authorization (Callable): A callback that receives method-specific information to continue login.
            All methods: 'credentials_source' parameter of type AwsCredentialsSource
            For Deadline Cloud monitor: No additional parameters
        on_cancellation_check (Callable): A callback that allows the operation to cancel before login completes
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.
    """
    credentials_source = get_credentials_source(config)
    if credentials_source == AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN:
        return _login_deadline_cloud_monitor(
            on_pending_authorization, on_cancellation_check, config
        )
    raise UnsupportedProfileTypeForLoginLogout(
        "Logging in is only supported for AWS Profiles created by Deadline Cloud monitor."
    )


@api.record_function_latency_telemetry_event()
def logout(config: Optional[ConfigParser] = None) -> str:
    """
    For AWS profiles created by Deadline Cloud monitor, logs out of Deadline Cloud.

     Args:
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.
    """
    credentials_source = get_credentials_source(config)
    if credentials_source == AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN:
        # Deadline Cloud monitor writes the absolute path to itself to the config file
        deadline_cloud_monitor_path = get_setting("deadline-cloud-monitor.path", config=config)
        profile_name = get_setting("defaults.aws_profile_name", config=config)
        args = [deadline_cloud_monitor_path, "logout", "--profile", profile_name]

        # Open Deadline Cloud monitor, blocking
        # Unlike login, that opens the regular Deadline Cloud monitor GUI, logout is a CLI command that clears the profile
        # This makes it easier as we can execute and look at the return cdoe
        try:
            output = subprocess.check_output(args)
        except FileNotFoundError:
            raise DeadlineOperationError(
                f"Could not find Deadline Cloud monitor at {deadline_cloud_monitor_path}. "
                f"Please ensure Deadline Cloud monitor is installed correctly and set up the {profile_name} profile again."
            )
        except subprocess.CalledProcessError as e:
            raise DeadlineOperationError(
                f"Deadline Cloud monitor was unable to log out the profile {profile_name}."
                f"Return code {e.returncode}: {e.output}"
            )

        # Force a refresh of the cached boto3 Session
        _session.invalidate_boto3_session_cache()
        return output.decode("utf8")
    raise UnsupportedProfileTypeForLoginLogout(
        "Logging out is only supported for AWS Profiles created by Deadline Cloud monitor."
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Contains methods for decoding and validating Asset Manifests."""

from __future__ import annotations

import json
import re
from typing import Any, Optional, Tuple

from ..exceptions import ManifestDecodeValidationError
from .base_manifest import BaseAssetManifest
from .manifest_model import ManifestModelRegistry
from .versions import ManifestVersion
from .v2023_03_03.validate import validate_manifest_2023_03_03

alphanum_regex = re.compile("[a-zA-Z0-9]+")


def validate_manifest(
    manifest: dict[str, Any], version: ManifestVersion
) -> Tuple[bool, Optional[str]]:
    """
    Checks if the given manifest is valid for the given manifest version. Returns True if the manifest
    is valid for the given version. Returns False and a string explaining the error if the manifest is not valid.
    """
    if version == ManifestVersion.v2023_03_03:
        return validate_manifest_2023_03_03(manifest)
    else:
        return False, f"Version {version} is not supported"


def decode_manifest(manifest: str) -> BaseAssetManifest:
    """
    Takes in a manifest string and returns an Asset Manifest object.
    A ManifestDecodeValidationError will be raised if the manifest version is unknown or
    the manifest is not valid.
    """
    document: dict[str, Any] = json.loads(manifest)

    try:
        version = ManifestVersion(document["manifestVersion"])
    except ValueError:
        # Value of the manifest version is not one we know.
        supported_versions = ", ".join(
            [v.value for v in ManifestVersion if v != ManifestVersion.UNDEFINED]
        )
        raise ManifestDecodeValidationError(
            f"Unknown manifest version: {document['manifestVersion']} "
            f"(Currently supported Manifest versions: {supported_versions})"
        )
    except KeyError:
        raise ManifestDecodeValidationError(
            'Manifest is missing the required "manifestVersion" field'
        )

    manifest_valid, error_string = validate_manifest(document, version)

    if not manifest_valid:
        raise ManifestDecodeValidationError(error_string)

    manifest_model = ManifestModelRegistry.get_manifest_model(version=version)
    decoded_manifest = manifest_model.AssetManifest.decode(manifest_data=document)

    # Validate hashes are alphanumeric
    for path in decoded_manifest.paths:
        if alphanum_regex.fullmatch(path.hash) is None:
            raise ManifestDecodeValidationError(
                f"The hash {path.hash} for path {path.path} is not alphanumeric"
            )

    return decoded_manifest
